# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools
import re
import warnings
from contextlib import suppress

from astropy.io.fits.card import Card
from astropy.io.fits.column import KEYWORD_NAMES as TABLE_KEYWORD_NAMES
from astropy.io.fits.column import TDEF_RE, ColDefs, Column
from astropy.io.fits.header import Header
from astropy.utils.exceptions import AstropyUserWarning

from .settings import (
    CMTYPE_ALIASES,
    COMPRESSION_TYPES,
    DEFAULT_BLOCK_SIZE,
    DEFAULT_BYTE_PIX,
    DEFAULT_COMPRESSION_TYPE,
    DEFAULT_DITHER_SEED,
    DEFAULT_HCOMP_SCALE,
    DEFAULT_HCOMP_SMOOTH,
    DEFAULT_QUANTIZE_LEVEL,
    DEFAULT_QUANTIZE_METHOD,
    NO_DITHER,
    QUANTIZE_METHOD_NAMES,
    SUBTRACTIVE_DITHER_1,
    SUBTRACTIVE_DITHER_2,
)
from .utils import _validate_tile_shape

__all__ = [
    "CompImageHeader",
    "_bintable_header_to_image_header",
    "_image_header_to_bintable_header_and_coldefs",
]


class CompImageHeader(Header):
    """
    Header object for compressed image HDUs designed to keep the compression
    header and the underlying image header properly synchronized.

    This essentially wraps the image header, so that all values are read from
    and written to the image header.  However, updates to the image header will
    also update the table header where appropriate.

    Note that if no image header is passed in, the code will instantiate a
    regular `~astropy.io.fits.Header`.
    """

    # TODO: The difficulty of implementing this screams a need to rewrite this
    # module

    _keyword_remaps = {
        "SIMPLE": "ZSIMPLE",
        "XTENSION": "ZTENSION",
        "BITPIX": "ZBITPIX",
        "NAXIS": "ZNAXIS",
        "EXTEND": "ZEXTEND",
        "BLOCKED": "ZBLOCKED",
        "PCOUNT": "ZPCOUNT",
        "GCOUNT": "ZGCOUNT",
        "CHECKSUM": "ZHECKSUM",
        "DATASUM": "ZDATASUM",
    }

    _zdef_re = re.compile(r"(?P<label>^[Zz][a-zA-Z]*)(?P<num>[1-9][0-9 ]*$)?")
    _compression_keywords = set(_keyword_remaps.values()).union(
        ["ZIMAGE", "ZCMPTYPE", "ZMASKCMP", "ZQUANTIZ", "ZDITHER0"]
    )
    _indexed_compression_keywords = {"ZNAXIS", "ZTILE", "ZNAME", "ZVAL"}
    # TODO: Once it place it should be possible to manage some of this through
    # the schema system, but it's not quite ready for that yet.  Also it still
    # makes more sense to change CompImageHDU to subclass ImageHDU :/

    def __new__(cls, table_header, image_header=None):
        # 2019-09-14 (MHvK): No point wrapping anything if no image_header is
        # given.  This happens if __getitem__ and copy are called - our super
        # class will aim to initialize a new, possibly partially filled
        # header, but we cannot usefully deal with that.
        # TODO: the above suggests strongly we should *not* subclass from
        # Header.  See also comment above about the need for reorganization.
        if image_header is None:
            return Header(table_header)
        else:
            return super().__new__(cls)

    def __init__(self, table_header, image_header):
        self._cards = image_header._cards
        self._keyword_indices = image_header._keyword_indices
        self._rvkc_indices = image_header._rvkc_indices
        self._modified = image_header._modified
        self._table_header = table_header

    # We need to override and Header methods that can modify the header, and
    # ensure that they sync with the underlying _table_header

    def __setitem__(self, key, value):
        # This isn't pretty, but if the `key` is either an int or a tuple we
        # need to figure out what keyword name that maps to before doing
        # anything else; these checks will be repeated later in the
        # super().__setitem__ call but I don't see another way around it
        # without some major refactoring
        if self._set_slice(key, value, self):
            return

        if isinstance(key, int):
            keyword, index = self._keyword_from_index(key)
        elif isinstance(key, tuple):
            keyword, index = key
        else:
            # We don't want to specify and index otherwise, because that will
            # break the behavior for new keywords and for commentary keywords
            keyword, index = key, None

        if self._is_reserved_keyword(keyword):
            return

        super().__setitem__(key, value)

        if index is not None:
            remapped_keyword = self._remap_keyword(keyword)
            self._table_header[remapped_keyword, index] = value
        # Else this will pass through to ._update

    def __delitem__(self, key):
        if isinstance(key, slice) or self._haswildcard(key):
            # If given a slice pass that on to the superclass and bail out
            # early; we only want to make updates to _table_header when given
            # a key specifying a single keyword
            return super().__delitem__(key)

        if isinstance(key, int):
            keyword, index = self._keyword_from_index(key)
        elif isinstance(key, tuple):
            keyword, index = key
        else:
            keyword, index = key, None

        if key not in self:
            raise KeyError(f"Keyword {key!r} not found.")

        super().__delitem__(key)

        remapped_keyword = self._remap_keyword(keyword)

        if remapped_keyword in self._table_header:
            if index is not None:
                del self._table_header[(remapped_keyword, index)]
            else:
                del self._table_header[remapped_keyword]

    def append(self, card=None, useblanks=True, bottom=False, end=False):
        # This logic unfortunately needs to be duplicated from the base class
        # in order to determine the keyword
        if isinstance(card, str):
            card = Card(card)
        elif isinstance(card, tuple):
            card = Card(*card)
        elif card is None:
            card = Card()
        elif not isinstance(card, Card):
            raise ValueError(
                "The value appended to a Header must be either a keyword or "
                f"(keyword, value, [comment]) tuple; got: {card!r}"
            )

        if self._is_reserved_keyword(card.keyword):
            return

        super().append(card=card, useblanks=useblanks, bottom=bottom, end=end)

        remapped_keyword = self._remap_keyword(card.keyword)

        # card.keyword strips the HIERARCH if present so this must be added
        # back to avoid a warning.
        if str(card).startswith("HIERARCH ") and not remapped_keyword.startswith(
            "HIERARCH "
        ):
            remapped_keyword = "HIERARCH " + remapped_keyword

        card = Card(remapped_keyword, card.value, card.comment)

        # Here we disable the use of blank cards, because the call above to
        # Header.append may have already deleted a blank card in the table
        # header, thanks to inheritance: Header.append calls 'del self[-1]'
        # to delete a blank card, which calls CompImageHeader.__deltitem__,
        # which deletes the blank card both in the image and the table headers!
        self._table_header.append(card=card, useblanks=False, bottom=bottom, end=end)

    def insert(self, key, card, useblanks=True, after=False):
        if isinstance(key, int):
            # Determine condition to pass through to append
            if after:
                if key == -1:
                    key = len(self._cards)
                else:
                    key += 1

            if key >= len(self._cards):
                self.append(card, end=True)
                return

        if isinstance(card, str):
            card = Card(card)
        elif isinstance(card, tuple):
            card = Card(*card)
        elif not isinstance(card, Card):
            raise ValueError(
                "The value inserted into a Header must be either a keyword or "
                f"(keyword, value, [comment]) tuple; got: {card!r}"
            )

        if self._is_reserved_keyword(card.keyword):
            return

        # Now the tricky part is to determine where to insert in the table
        # header.  If given a numerical index we need to map that to the
        # corresponding index in the table header.  Although rare, there may be
        # cases where there is no mapping in which case we just try the same
        # index
        # NOTE: It is crucial that remapped_index in particular is figured out
        # before the image header is modified
        remapped_index = self._remap_index(key)
        remapped_keyword = self._remap_keyword(card.keyword)

        super().insert(key, card, useblanks=useblanks, after=after)

        card = Card(remapped_keyword, card.value, card.comment)

        # Here we disable the use of blank cards, because the call above to
        # Header.insert may have already deleted a blank card in the table
        # header, thanks to inheritance: Header.insert calls 'del self[-1]'
        # to delete a blank card, which calls CompImageHeader.__delitem__,
        # which deletes the blank card both in the image and the table headers!
        self._table_header.insert(remapped_index, card, useblanks=False, after=after)

    def _update(self, card):
        keyword = card[0]

        if self._is_reserved_keyword(keyword):
            return

        super()._update(card)

        if keyword in Card._commentary_keywords:
            # Otherwise this will result in a duplicate insertion
            return

        remapped_keyword = self._remap_keyword(keyword)
        self._table_header._update((remapped_keyword,) + card[1:])

    # Last piece needed (I think) for synchronizing with the real header
    # This one is tricky since _relativeinsert calls insert
    def _relativeinsert(self, card, before=None, after=None, replace=False):
        keyword = card[0]

        if self._is_reserved_keyword(keyword):
            return

        # Now we have to figure out how to remap 'before' and 'after'
        if before is None:
            if isinstance(after, int):
                remapped_after = self._remap_index(after)
            else:
                remapped_after = self._remap_keyword(after)
            remapped_before = None
        else:
            if isinstance(before, int):
                remapped_before = self._remap_index(before)
            else:
                remapped_before = self._remap_keyword(before)
            remapped_after = None

        super()._relativeinsert(card, before=before, after=after, replace=replace)

        remapped_keyword = self._remap_keyword(keyword)

        card = Card(remapped_keyword, card[1], card[2])
        self._table_header._relativeinsert(
            card, before=remapped_before, after=remapped_after, replace=replace
        )

    @classmethod
    def _is_reserved_keyword(cls, keyword, warn=True):
        msg = (
            f"Keyword {keyword!r} is reserved for use by the FITS Tiled Image "
            "Convention and will not be stored in the header for the "
            "image being compressed."
        )

        if keyword == "TFIELDS":
            if warn:
                warnings.warn(msg)
            return True

        m = TDEF_RE.match(keyword)

        if m and m.group("label").upper() in TABLE_KEYWORD_NAMES:
            if warn:
                warnings.warn(msg)
            return True

        m = cls._zdef_re.match(keyword)

        if m:
            label = m.group("label").upper()
            num = m.group("num")
            if num is not None and label in cls._indexed_compression_keywords:
                if warn:
                    warnings.warn(msg)
                return True
            elif label in cls._compression_keywords:
                if warn:
                    warnings.warn(msg)
                return True

        return False

    @classmethod
    def _remap_keyword(cls, keyword):
        # Given a keyword that one might set on an image, remap that keyword to
        # the name used for it in the COMPRESSED HDU header
        # This is mostly just a lookup in _keyword_remaps, but needs handling
        # for NAXISn keywords

        is_naxisn = False
        if keyword[:5] == "NAXIS":
            with suppress(ValueError):
                index = int(keyword[5:])
                is_naxisn = index > 0

        if is_naxisn:
            return f"ZNAXIS{index}"

        # If the keyword does not need to be remapped then just return the
        # original keyword
        return cls._keyword_remaps.get(keyword, keyword)

    def _remap_index(self, idx):
        # Given an integer index into this header, map that to the index in the
        # table header for the same card.  If the card doesn't exist in the
        # table header (generally should *not* be the case) this will just
        # return the same index
        # This *does* also accept a keyword or (keyword, repeat) tuple and
        # obtains the associated numerical index with self._cardindex
        if not isinstance(idx, int):
            idx = self._cardindex(idx)

        keyword, repeat = self._keyword_from_index(idx)
        remapped_insert_keyword = self._remap_keyword(keyword)

        with suppress(IndexError, KeyError):
            idx = self._table_header._cardindex((remapped_insert_keyword, repeat))

        return idx

    def clear(self):
        """
        Remove all cards from the header.
        """
        self._table_header.clear()
        super().clear()


def _bintable_header_to_image_header(bintable_header):
    # Start with a copy of the table header.
    image_header = bintable_header.copy()

    # Delete cards that are related to the table.  And move
    # the values of those cards that relate to the image from
    # their corresponding table cards.  These include
    # ZBITPIX -> BITPIX, ZNAXIS -> NAXIS, and ZNAXISn -> NAXISn.
    # (Note: Used set here instead of list in case there are any duplicate
    # keywords, which there may be in some pathological cases:
    # https://github.com/astropy/astropy/issues/2750
    for keyword in set(image_header):
        if CompImageHeader._is_reserved_keyword(keyword, warn=False):
            del image_header[keyword]

    hcomments = bintable_header.comments

    if "ZSIMPLE" in bintable_header:
        image_header.set(
            "SIMPLE", bintable_header["ZSIMPLE"], hcomments["ZSIMPLE"], before=0
        )
        del image_header["XTENSION"]
    elif "ZTENSION" in bintable_header:
        if bintable_header["ZTENSION"] != "IMAGE":
            warnings.warn(
                "ZTENSION keyword in compressed extension != 'IMAGE'",
                AstropyUserWarning,
            )
        image_header.set("XTENSION", "IMAGE", hcomments["ZTENSION"], before=0)
    else:
        image_header.set("XTENSION", "IMAGE", before=0)

    image_header.set(
        "BITPIX", bintable_header["ZBITPIX"], hcomments["ZBITPIX"], before=1
    )

    image_header.set("NAXIS", bintable_header["ZNAXIS"], hcomments["ZNAXIS"], before=2)

    last_naxis = "NAXIS"
    for idx in range(image_header["NAXIS"]):
        znaxis = "ZNAXIS" + str(idx + 1)
        naxis = znaxis[1:]
        image_header.set(
            naxis, bintable_header[znaxis], hcomments[znaxis], after=last_naxis
        )
        last_naxis = naxis

    # Delete any other spurious NAXISn keywords:
    naxis = image_header["NAXIS"]
    for keyword in list(image_header["NAXIS?*"]):
        try:
            n = int(keyword[5:])
        except Exception:
            continue

        if n > naxis:
            del image_header[keyword]

    # Although PCOUNT and GCOUNT are considered mandatory for IMAGE HDUs,
    # ZPCOUNT and ZGCOUNT are optional, probably because for IMAGE HDUs
    # their values are always 0 and 1 respectively
    if "ZPCOUNT" in bintable_header:
        image_header.set(
            "PCOUNT",
            bintable_header["ZPCOUNT"],
            hcomments["ZPCOUNT"],
            after=last_naxis,
        )
    else:
        image_header.set("PCOUNT", 0, after=last_naxis)

    if "ZGCOUNT" in bintable_header:
        image_header.set(
            "GCOUNT", bintable_header["ZGCOUNT"], hcomments["ZGCOUNT"], after="PCOUNT"
        )
    else:
        image_header.set("GCOUNT", 1, after="PCOUNT")

    if "ZEXTEND" in bintable_header:
        image_header.set("EXTEND", bintable_header["ZEXTEND"], hcomments["ZEXTEND"])

    if "ZBLOCKED" in bintable_header:
        image_header.set("BLOCKED", bintable_header["ZBLOCKED"], hcomments["ZBLOCKED"])

    # Move the ZHECKSUM and ZDATASUM cards to the image header
    # as CHECKSUM and DATASUM
    if "ZHECKSUM" in bintable_header:
        image_header.set("CHECKSUM", bintable_header["ZHECKSUM"], hcomments["ZHECKSUM"])

    if "ZDATASUM" in bintable_header:
        image_header.set("DATASUM", bintable_header["ZDATASUM"], hcomments["ZDATASUM"])

    # Remove the EXTNAME card if the value in the table header
    # is the default value of COMPRESSED_IMAGE.
    if "EXTNAME" in image_header and image_header["EXTNAME"] == "COMPRESSED_IMAGE":
        del image_header["EXTNAME"]

    # Remove the PCOUNT GCOUNT cards if the uncompressed header is
    # from a primary HDU
    if "SIMPLE" in image_header:
        del image_header["PCOUNT"]
        del image_header["GCOUNT"]

    # Look to see if there are any blank cards in the table
    # header.  If there are, there should be the same number
    # of blank cards in the image header.  Add blank cards to
    # the image header to make it so.
    table_blanks = bintable_header._countblanks()
    image_blanks = image_header._countblanks()

    for _ in range(table_blanks - image_blanks):
        image_header.append()

    # Create the CompImageHeader that syncs with the table header
    return CompImageHeader(bintable_header, image_header)


def _image_header_to_bintable_header_and_coldefs(
    image_header,
    generated_image_header,
    bintable_header,
    name=None,
    huge_hdu=False,
    compression_type=None,
    tile_shape=None,
    hcomp_scale=None,
    hcomp_smooth=None,
    quantize_level=None,
    quantize_method=None,
    dither_seed=None,
    axes=None,
    generate_dither_seed=None,
):
    # NOTE: image_header is the header that a user would see as the image
    # header which they might have set things like BSCALE and BZERO on, or
    # added history or comments to. Whereas generated_image_header is the
    # image header as converted/generated from the existing binary table HDU.

    # Update the extension name in the table header
    if not name and "EXTNAME" not in bintable_header:
        # Do not sync this with the image header since the default
        # name is specific to the table header.
        bintable_header.set(
            "EXTNAME",
            "COMPRESSED_IMAGE",
            "name of this binary table extension",
            after="TFIELDS",
        )

    # Set the compression type in the table header.
    if compression_type:
        if compression_type not in COMPRESSION_TYPES:
            warnings.warn(
                "Unknown compression type provided (supported are {}). "
                "Default ({}) compression will be used.".format(
                    ", ".join(map(repr, COMPRESSION_TYPES)),
                    DEFAULT_COMPRESSION_TYPE,
                ),
                AstropyUserWarning,
            )
            compression_type = DEFAULT_COMPRESSION_TYPE

        bintable_header.set(
            "ZCMPTYPE", compression_type, "compression algorithm", after="TFIELDS"
        )
    else:
        compression_type = CMTYPE_ALIASES.get(compression_type, compression_type)

    # If the input image header had BSCALE/BZERO cards, then insert
    # them in the table header.

    if image_header:
        bzero = image_header.get("BZERO", 0.0)
        bscale = image_header.get("BSCALE", 1.0)
        after_keyword = "EXTNAME"

        if bscale != 1.0:
            bintable_header.set("BSCALE", bscale, after=after_keyword)
            after_keyword = "BSCALE"

        if bzero != 0.0:
            bintable_header.set("BZERO", bzero, after=after_keyword)

    try:
        bitpix_comment = image_header.comments["BITPIX"]
    except (AttributeError, KeyError):
        bitpix_comment = "data type of original image"

    try:
        naxis_comment = image_header.comments["NAXIS"]
    except (AttributeError, KeyError):
        naxis_comment = "dimension of original image"

    # Set the label for the first column in the table

    bintable_header.set(
        "TTYPE1", "COMPRESSED_DATA", "label for field 1", after="TFIELDS"
    )

    # Set the data format for the first column.  It is dependent
    # on the requested compression type.

    if compression_type == "PLIO_1":
        tform1 = "1QI" if huge_hdu else "1PI"
    else:
        tform1 = "1QB" if huge_hdu else "1PB"

    bintable_header.set(
        "TFORM1",
        tform1,
        "data format of field: variable length array",
        after="TTYPE1",
    )

    # Create the first column for the table.  This column holds the
    # compressed data.
    col1 = Column(name=bintable_header["TTYPE1"], format=tform1)

    # Create the additional columns required for floating point
    # data and calculate the width of the output table.

    zbitpix = generated_image_header["BITPIX"]

    if zbitpix < 0 and quantize_level != 0.0:
        # floating point image has 'COMPRESSED_DATA',
        # 'UNCOMPRESSED_DATA', 'ZSCALE', and 'ZZERO' columns (unless using
        # lossless compression, per CFITSIO)
        ncols = 4

        # CFITSIO 3.28 and up automatically use the GZIP_COMPRESSED_DATA
        # store floating point data that couldn't be quantized, instead
        # of the UNCOMPRESSED_DATA column.  There's no way to control
        # this behavior so the only way to determine which behavior will
        # be employed is via the CFITSIO version

        ttype2 = "GZIP_COMPRESSED_DATA"
        # The required format for the GZIP_COMPRESSED_DATA is actually
        # missing from the standard docs, but CFITSIO suggests it
        # should be 1PB, which is logical.
        tform2 = "1QB" if huge_hdu else "1PB"

        # Set up the second column for the table that will hold any
        # uncompressable data.
        bintable_header.set("TTYPE2", ttype2, "label for field 2", after="TFORM1")

        bintable_header.set(
            "TFORM2",
            tform2,
            "data format of field: variable length array",
            after="TTYPE2",
        )

        col2 = Column(name=ttype2, format=tform2)

        # Set up the third column for the table that will hold
        # the scale values for quantized data.
        bintable_header.set("TTYPE3", "ZSCALE", "label for field 3", after="TFORM2")
        bintable_header.set(
            "TFORM3", "1D", "data format of field: 8-byte DOUBLE", after="TTYPE3"
        )
        col3 = Column(name=bintable_header["TTYPE3"], format=bintable_header["TFORM3"])

        # Set up the fourth column for the table that will hold
        # the zero values for the quantized data.
        bintable_header.set("TTYPE4", "ZZERO", "label for field 4", after="TFORM3")
        bintable_header.set(
            "TFORM4", "1D", "data format of field: 8-byte DOUBLE", after="TTYPE4"
        )
        after = "TFORM4"
        col4 = Column(name=bintable_header["TTYPE4"], format=bintable_header["TFORM4"])

        # Create the ColDefs object for the table
        cols = ColDefs([col1, col2, col3, col4])
    else:
        # default table has just one 'COMPRESSED_DATA' column
        ncols = 1
        after = "TFORM1"

        # remove any header cards for the additional columns that
        # may be left over from the previous data
        to_remove = ["TTYPE2", "TFORM2", "TTYPE3", "TFORM3", "TTYPE4", "TFORM4"]

        for k in to_remove:
            try:
                del bintable_header[k]
            except KeyError:
                pass

        # Create the ColDefs object for the table
        cols = ColDefs([col1])

    # Update the table header with the width of the table, the
    # number of fields in the table, the indicator for a compressed
    # image HDU, the data type of the image data and the number of
    # dimensions in the image data array.
    bintable_header.set("NAXIS1", cols.dtype.itemsize, "width of table in bytes")
    bintable_header.set(
        "TFIELDS", ncols, "number of fields in each row", after="GCOUNT"
    )
    bintable_header.set(
        "ZIMAGE", True, "extension contains compressed image", after=after
    )
    bintable_header.set("ZBITPIX", zbitpix, bitpix_comment, after="ZIMAGE")
    bintable_header.set(
        "ZNAXIS", generated_image_header["NAXIS"], naxis_comment, after="ZBITPIX"
    )

    # Strip the table header of all the ZNAZISn and ZTILEn keywords
    # that may be left over from the previous data

    for idx in itertools.count(1):
        try:
            del bintable_header["ZNAXIS" + str(idx)]
            del bintable_header["ZTILE" + str(idx)]
        except KeyError:
            break

    # Verify that any input tile size parameter is the appropriate
    # size to match the HDU's data.

    tile_shape = _validate_tile_shape(
        tile_shape=tile_shape,
        compression_type=compression_type,
        image_header=generated_image_header,
    )

    # Set up locations for writing the next cards in the header.
    last_znaxis = "ZNAXIS"

    if generated_image_header["NAXIS"] > 0:
        after1 = "ZNAXIS1"
    else:
        after1 = "ZNAXIS"

    # Calculate the number of rows in the output table and
    # write the ZNAXISn and ZTILEn cards to the table header.
    nrows = 0

    for idx, axis in enumerate(axes):
        naxis = "NAXIS" + str(idx + 1)
        znaxis = "ZNAXIS" + str(idx + 1)
        ztile = "ZTILE" + str(idx + 1)

        ts = tile_shape[len(axes) - 1 - idx]

        if not nrows:
            nrows = (axis - 1) // ts + 1
        else:
            nrows *= (axis - 1) // ts + 1

        if image_header and naxis in image_header:
            bintable_header.set(
                znaxis, axis, image_header.comments[naxis], after=last_znaxis
            )
        else:
            bintable_header.set(
                znaxis, axis, "length of original image axis", after=last_znaxis
            )

        bintable_header.set(ztile, ts, "size of tiles to be compressed", after=after1)
        last_znaxis = znaxis
        after1 = ztile

    # Set the NAXIS2 header card in the table hdu to the number of
    # rows in the table.
    bintable_header.set("NAXIS2", nrows, "number of rows in table")

    # Set the compression parameters in the table header.

    # First, setup the values to be used for the compression parameters
    # in case none were passed in.  This will be either the value
    # already in the table header for that parameter or the default
    # value.
    for idx in itertools.count(1):
        zname = "ZNAME" + str(idx)
        if zname not in bintable_header:
            break
        zval = "ZVAL" + str(idx)
        if bintable_header[zname] == "NOISEBIT":
            if quantize_level is None:
                quantize_level = bintable_header[zval]
        if bintable_header[zname] == "SCALE   ":
            if hcomp_scale is None:
                hcomp_scale = bintable_header[zval]
        if bintable_header[zname] == "SMOOTH  ":
            if hcomp_smooth is None:
                hcomp_smooth = bintable_header[zval]

    if quantize_level is None:
        quantize_level = DEFAULT_QUANTIZE_LEVEL

    if hcomp_scale is None:
        hcomp_scale = DEFAULT_HCOMP_SCALE

    if hcomp_smooth is None:
        hcomp_smooth = DEFAULT_HCOMP_SMOOTH

    # Next, strip the table header of all the ZNAMEn and ZVALn keywords
    # that may be left over from the previous data
    for idx in itertools.count(1):
        zname = "ZNAME" + str(idx)
        if zname not in bintable_header:
            break
        zval = "ZVAL" + str(idx)
        del bintable_header[zname]
        del bintable_header[zval]

    # Finally, put the appropriate keywords back based on the
    # compression type.

    after_keyword = "ZCMPTYPE"
    idx = 1

    if compression_type == "RICE_1":
        bintable_header.set(
            "ZNAME1", "BLOCKSIZE", "compression block size", after=after_keyword
        )
        bintable_header.set(
            "ZVAL1", DEFAULT_BLOCK_SIZE, "pixels per block", after="ZNAME1"
        )

        bintable_header.set(
            "ZNAME2", "BYTEPIX", "bytes per pixel (1, 2, 4, or 8)", after="ZVAL1"
        )

        if bintable_header["ZBITPIX"] == 8:
            bytepix = 1
        elif bintable_header["ZBITPIX"] == 16:
            bytepix = 2
        else:
            bytepix = DEFAULT_BYTE_PIX

        bintable_header.set(
            "ZVAL2", bytepix, "bytes per pixel (1, 2, 4, or 8)", after="ZNAME2"
        )
        after_keyword = "ZVAL2"
        idx = 3
    elif compression_type == "HCOMPRESS_1":
        bintable_header.set(
            "ZNAME1", "SCALE", "HCOMPRESS scale factor", after=after_keyword
        )
        bintable_header.set(
            "ZVAL1", hcomp_scale, "HCOMPRESS scale factor", after="ZNAME1"
        )
        bintable_header.set(
            "ZNAME2", "SMOOTH", "HCOMPRESS smooth option", after="ZVAL1"
        )
        bintable_header.set(
            "ZVAL2", hcomp_smooth, "HCOMPRESS smooth option", after="ZNAME2"
        )
        after_keyword = "ZVAL2"
        idx = 3

    if generated_image_header["BITPIX"] < 0:  # floating point image
        bintable_header.set(
            "ZNAME" + str(idx),
            "NOISEBIT",
            "floating point quantization level",
            after=after_keyword,
        )
        bintable_header.set(
            "ZVAL" + str(idx),
            quantize_level,
            "floating point quantization level",
            after="ZNAME" + str(idx),
        )

        # Add the dither method and seed
        if quantize_method:
            if quantize_method not in [
                NO_DITHER,
                SUBTRACTIVE_DITHER_1,
                SUBTRACTIVE_DITHER_2,
            ]:
                name = QUANTIZE_METHOD_NAMES[DEFAULT_QUANTIZE_METHOD]
                warnings.warn(
                    "Unknown quantization method provided.  "
                    f"Default method ({name}) used."
                )
                quantize_method = DEFAULT_QUANTIZE_METHOD

            if quantize_method == NO_DITHER:
                zquantiz_comment = "No dithering during quantization"
            else:
                zquantiz_comment = "Pixel Quantization Algorithm"

            bintable_header.set(
                "ZQUANTIZ",
                QUANTIZE_METHOD_NAMES[quantize_method],
                zquantiz_comment,
                after="ZVAL" + str(idx),
            )
        else:
            # If the ZQUANTIZ keyword is missing the default is to assume
            # no dithering, rather than whatever DEFAULT_QUANTIZE_METHOD
            # is set to
            quantize_method = bintable_header.get("ZQUANTIZ", NO_DITHER)

            if isinstance(quantize_method, str):
                for k, v in QUANTIZE_METHOD_NAMES.items():
                    if v.upper() == quantize_method:
                        quantize_method = k
                        break
                else:
                    quantize_method = NO_DITHER

        if quantize_method == NO_DITHER:
            if "ZDITHER0" in bintable_header:
                # If dithering isn't being used then there's no reason to
                # keep the ZDITHER0 keyword
                del bintable_header["ZDITHER0"]
        else:
            if dither_seed:
                dither_seed = generate_dither_seed(dither_seed)
            elif "ZDITHER0" in bintable_header:
                dither_seed = bintable_header["ZDITHER0"]
            else:
                dither_seed = generate_dither_seed(DEFAULT_DITHER_SEED)

            bintable_header.set(
                "ZDITHER0",
                dither_seed,
                "dithering offset when quantizing floats",
                after="ZQUANTIZ",
            )

    if image_header:
        # Move SIMPLE card from the image header to the
        # table header as ZSIMPLE card.

        if "SIMPLE" in image_header:
            bintable_header.set(
                "ZSIMPLE",
                image_header["SIMPLE"],
                image_header.comments["SIMPLE"],
                before="ZBITPIX",
            )

        # Move EXTEND card from the image header to the
        # table header as ZEXTEND card.

        if "EXTEND" in image_header:
            bintable_header.set(
                "ZEXTEND", image_header["EXTEND"], image_header.comments["EXTEND"]
            )

        # Move BLOCKED card from the image header to the
        # table header as ZBLOCKED card.

        if "BLOCKED" in image_header:
            bintable_header.set(
                "ZBLOCKED",
                image_header["BLOCKED"],
                image_header.comments["BLOCKED"],
            )

        # Move XTENSION card from the image header to the
        # table header as ZTENSION card.

        # Since we only handle compressed IMAGEs, ZTENSION should
        # always be IMAGE, even if the caller has passed in a header
        # for some other type of extension.
        if "XTENSION" in image_header:
            bintable_header.set(
                "ZTENSION",
                "IMAGE",
                image_header.comments["XTENSION"],
                before="ZBITPIX",
            )

        # Move PCOUNT and GCOUNT cards from image header to the table
        # header as ZPCOUNT and ZGCOUNT cards.

        if "PCOUNT" in image_header:
            bintable_header.set(
                "ZPCOUNT",
                image_header["PCOUNT"],
                image_header.comments["PCOUNT"],
                after=last_znaxis,
            )

        if "GCOUNT" in image_header:
            bintable_header.set(
                "ZGCOUNT",
                image_header["GCOUNT"],
                image_header.comments["GCOUNT"],
                after="ZPCOUNT",
            )

        # Move CHECKSUM and DATASUM cards from the image header to the
        # table header as XHECKSUM and XDATASUM cards.

        if "CHECKSUM" in image_header:
            bintable_header.set(
                "ZHECKSUM",
                image_header["CHECKSUM"],
                image_header.comments["CHECKSUM"],
            )

        if "DATASUM" in image_header:
            bintable_header.set(
                "ZDATASUM",
                image_header["DATASUM"],
                image_header.comments["DATASUM"],
            )
    else:
        # Move XTENSION card from the image header to the
        # table header as ZTENSION card.

        # Since we only handle compressed IMAGEs, ZTENSION should
        # always be IMAGE, even if the caller has passed in a header
        # for some other type of extension.
        if "XTENSION" in generated_image_header:
            bintable_header.set(
                "ZTENSION",
                "IMAGE",
                generated_image_header.comments["XTENSION"],
                before="ZBITPIX",
            )

        # Move PCOUNT and GCOUNT cards from image header to the table
        # header as ZPCOUNT and ZGCOUNT cards.

        if "PCOUNT" in generated_image_header:
            bintable_header.set(
                "ZPCOUNT",
                generated_image_header["PCOUNT"],
                generated_image_header.comments["PCOUNT"],
                after=last_znaxis,
            )

        if "GCOUNT" in generated_image_header:
            bintable_header.set(
                "ZGCOUNT",
                generated_image_header["GCOUNT"],
                generated_image_header.comments["GCOUNT"],
                after="ZPCOUNT",
            )

    # When we have an image checksum we need to ensure that the same
    # number of blank cards exist in the table header as there were in
    # the image header.  This allows those blank cards to be carried
    # over to the image header when the hdu is uncompressed.

    if "ZHECKSUM" in bintable_header:
        required_blanks = image_header._countblanks()
        image_blanks = generated_image_header._countblanks()
        table_blanks = bintable_header._countblanks()

        for _ in range(required_blanks - image_blanks):
            generated_image_header.append()
            table_blanks += 1

        for _ in range(required_blanks - table_blanks):
            bintable_header.append()

    return bintable_header, cols



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
from contextlib import nullcontext

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.contour import QuadContourSet
from packaging.version import Version

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.visualization.wcsaxes.core import WCSAxes
from astropy.visualization.wcsaxes.frame import (
    EllipticalFrame,
    RectangularFrame,
    RectangularFrame1D,
)
from astropy.visualization.wcsaxes.transforms import CurvedTransform
from astropy.visualization.wcsaxes.utils import get_coord_meta
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSWrapper, SlicedLowLevelWCS

ft_version = Version(mpl.ft2font.__freetype_version__)
FREETYPE_261 = ft_version == Version("2.6.1")

# We cannot use matplotlib.checkdep_usetex() anymore, see
# https://github.com/matplotlib/matplotlib/issues/23244
TEX_UNAVAILABLE = True

MATPLOTLIB_LT_3_7 = Version(mpl.__version__) < Version("3.7")


def teardown_function(function):
    plt.close("all")


def test_grid_regression(ignore_matplotlibrc):
    # Regression test for a bug that meant that if the rc parameter
    # axes.grid was set to True, WCSAxes would crash upon initialization.
    plt.rc("axes", grid=True)
    fig = plt.figure(figsize=(3, 3))
    WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])


def test_format_coord_regression(ignore_matplotlibrc, tmp_path):
    # Regression test for a bug that meant that if format_coord was called by
    # Matplotlib before the axes were drawn, an error occurred.
    fig = plt.figure(figsize=(3, 3))
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)
    assert ax.format_coord(10, 10) == ""
    assert ax.coords[0].format_coord(10) == ""
    assert ax.coords[1].format_coord(10) == ""
    fig.savefig(tmp_path / "nothing")
    assert ax.format_coord(10, 10) == "10.0 10.0 (world)"
    assert ax.coords[0].format_coord(10) == "10.0"
    assert ax.coords[1].format_coord(10) == "10.0"


TARGET_HEADER = fits.Header.fromstring(
    """
NAXIS   =                    2
NAXIS1  =                  200
NAXIS2  =                  100
CTYPE1  = 'RA---MOL'
CRPIX1  =                  500
CRVAL1  =                180.0
CDELT1  =                 -0.4
CUNIT1  = 'deg     '
CTYPE2  = 'DEC--MOL'
CRPIX2  =                  400
CRVAL2  =                  0.0
CDELT2  =                  0.4
CUNIT2  = 'deg     '
COORDSYS= 'icrs    '
""",
    sep="\n",
)


@pytest.mark.parametrize("grid_type", ["lines", "contours"])
def test_no_numpy_warnings(ignore_matplotlibrc, tmp_path, grid_type):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS(TARGET_HEADER))
    ax.imshow(np.zeros((100, 200)))
    ax.coords.grid(color="white", grid_type=grid_type)

    # There should be no warnings raised if some pixels are outside WCS
    # (since this is normal).
    # BUT our own catch_warning was ignoring some warnings before, so now we
    # have to catch it. Otherwise, the pytest filterwarnings=error
    # setting in pyproject.toml will fail this test.
    # There are actually multiple warnings but they are all similar.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=r".*converting a masked element to nan.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*No contour levels were found within the data range.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*np\.asscalar\(a\) is deprecated since NumPy v1\.16.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*PY_SSIZE_T_CLEAN will be required.*"
        )
        fig.savefig(tmp_path / "test.png")


def test_invalid_frame_overlay(ignore_matplotlibrc):
    # Make sure a nice error is returned if a frame doesn't exist
    ax = plt.subplot(1, 1, 1, projection=WCS(TARGET_HEADER))
    with pytest.raises(ValueError, match=r"Frame banana not found"):
        ax.get_coords_overlay("banana")

    with pytest.raises(ValueError, match=r"Unknown frame: banana"):
        get_coord_meta("banana")


def test_plot_coord_transform(ignore_matplotlibrc):
    twoMASS_k_header = get_pkg_data_filename("data/2MASS_k_header")
    twoMASS_k_header = fits.Header.fromtextfile(twoMASS_k_header)
    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_axes(
        [0.15, 0.15, 0.8, 0.8], projection=WCS(twoMASS_k_header), aspect="equal"
    )
    ax.set_xlim(-0.5, 720.5)
    ax.set_ylim(-0.5, 720.5)

    c = SkyCoord(359.76045223 * u.deg, 0.26876217 * u.deg)
    with pytest.raises(TypeError):
        ax.plot_coord(c, "o", transform=ax.get_transform("galactic"))


def test_scatter_coord_transform(ignore_matplotlibrc):
    twoMASS_k_header = get_pkg_data_filename("data/2MASS_k_header")
    twoMASS_k_header = fits.Header.fromtextfile(twoMASS_k_header)
    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_axes(
        [0.15, 0.15, 0.8, 0.8], projection=WCS(twoMASS_k_header), aspect="equal"
    )
    ax.set_xlim(-0.5, 720.5)
    ax.set_ylim(-0.5, 720.5)

    c = SkyCoord(359.76045223 * u.deg, 0.26876217 * u.deg)
    with pytest.raises(TypeError):
        ax.scatter_coord(c, marker="o", transform=ax.get_transform("galactic"))


def test_set_label_properties(ignore_matplotlibrc):
    # Regression test to make sure that arguments passed to
    # set_xlabel/set_ylabel are passed to the underlying coordinate helpers

    ax = plt.subplot(1, 1, 1, projection=WCS(TARGET_HEADER))

    ax.set_xlabel("Test x label", labelpad=2, color="red")
    ax.set_ylabel("Test y label", labelpad=3, color="green")

    assert ax.coords[0].axislabels.get_text() == "Test x label"
    assert ax.coords[0].axislabels.get_minpad("b") == 2
    assert ax.coords[0].axislabels.get_color() == "red"

    assert ax.coords[1].axislabels.get_text() == "Test y label"
    assert ax.coords[1].axislabels.get_minpad("l") == 3
    assert ax.coords[1].axislabels.get_color() == "green"

    assert ax.get_xlabel() == "Test x label"
    assert ax.get_ylabel() == "Test y label"


GAL_HEADER = fits.Header.fromstring(
    """
SIMPLE  =                    T / conforms to FITS standard
BITPIX  =                  -32 / array data type
NAXIS   =                    3 / number of array dimensions
NAXIS1  =                   31
NAXIS2  =                 2881
NAXIS3  =                  480
EXTEND  =                    T
CTYPE1  = 'DISTMOD '
CRVAL1  =                  3.5
CDELT1  =                  0.5
CRPIX1  =                  1.0
CTYPE2  = 'GLON-CAR'
CRVAL2  =                180.0
CDELT2  =               -0.125
CRPIX2  =                  1.0
CTYPE3  = 'GLAT-CAR'
CRVAL3  =                  0.0
CDELT3  =                0.125
CRPIX3  =                241.0
""",
    sep="\n",
)


def test_slicing_warnings(ignore_matplotlibrc, tmp_path):
    # Regression test to make sure that no warnings are emitted by the tick
    # locator for the sliced axis when slicing a cube.

    # Scalar case

    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]
    wcs3d.wcs.cunit = ["deg", "deg", "km/s"]
    wcs3d.wcs.crpix = [614.5, 856.5, 333]
    wcs3d.wcs.cdelt = [6.25, 6.25, 23]
    wcs3d.wcs.crval = [0.0, 0.0, 1.0]

    with warnings.catch_warnings():
        # https://github.com/astropy/astropy/issues/9690
        warnings.filterwarnings("ignore", message=r".*PY_SSIZE_T_CLEAN.*")
        plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))
        plt.savefig(tmp_path / "test.png")

    # Angle case

    wcs3d = WCS(GAL_HEADER)

    with warnings.catch_warnings():
        # https://github.com/astropy/astropy/issues/9690
        warnings.filterwarnings("ignore", message=r".*PY_SSIZE_T_CLEAN.*")
        plt.clf()
        plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 2))
        plt.savefig(tmp_path / "test.png")


def test_plt_xlabel_ylabel(tmp_path):
    # Regression test for a bug that happened when using plt.xlabel
    # and plt.ylabel with Matplotlib 3.0

    plt.subplot(projection=WCS())
    plt.xlabel("Galactic Longitude")
    plt.ylabel("Galactic Latitude")
    plt.savefig(tmp_path / "test.png")


def test_grid_type_contours_transform(tmp_path):
    # Regression test for a bug that caused grid_type='contours' to not work
    # with custom transforms

    class CustomTransform(CurvedTransform):
        # We deliberately don't define the inverse, and has_inverse should
        # default to False.

        def transform(self, values):
            return values * 1.3

    transform = CustomTransform()
    coord_meta = {
        "type": ("scalar", "scalar"),
        "unit": (u.m, u.s),
        "wrap": (None, None),
        "name": ("x", "y"),
    }

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], transform=transform, coord_meta=coord_meta)
    fig.add_axes(ax)
    ax.grid(grid_type="contours")
    fig.savefig(tmp_path / "test.png")


def test_plt_imshow_origin():
    # Regression test for a bug that caused origin to be set to upper when
    # plt.imshow was called.

    ax = plt.subplot(projection=WCS())
    plt.imshow(np.ones((2, 2)))
    assert ax.get_xlim() == (-0.5, 1.5)
    assert ax.get_ylim() == (-0.5, 1.5)


def test_ax_imshow_origin():
    # Regression test for a bug that caused origin to be set to upper when
    # ax.imshow was called with no origin

    ax = plt.subplot(projection=WCS())
    ax.imshow(np.ones((2, 2)))
    assert ax.get_xlim() == (-0.5, 1.5)
    assert ax.get_ylim() == (-0.5, 1.5)


def test_grid_contour_large_spacing(tmp_path):
    # Regression test for a bug that caused a crash when grid was called and
    # didn't produce grid lines (due e.g. to too large spacing) and was then
    # called again.

    filename = tmp_path / "test.png"

    ax = plt.subplot(projection=WCS())
    ax.set_xlim(-0.5, 1.5)
    ax.set_ylim(-0.5, 1.5)
    ax.coords[0].set_ticks(values=[] * u.one)

    ax.coords[0].grid(grid_type="contours")
    plt.savefig(filename)

    ax.coords[0].grid(grid_type="contours")
    plt.savefig(filename)


def test_contour_return():
    # Regression test for a bug that caused contour and contourf to return None
    # instead of the contour object.

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)

    cset = ax.contour(np.arange(16).reshape(4, 4), transform=ax.get_transform("world"))
    assert isinstance(cset, QuadContourSet)

    cset = ax.contourf(np.arange(16).reshape(4, 4), transform=ax.get_transform("world"))
    assert isinstance(cset, QuadContourSet)


def test_contour_empty():
    # Regression test for a bug that caused contour to crash if no contours
    # were present.

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)

    if MATPLOTLIB_LT_3_7:
        ctx = pytest.warns(
            UserWarning, match="No contour levels were found within the data range"
        )
    else:
        ctx = nullcontext()

    with ctx:
        ax.contour(np.zeros((4, 4)), transform=ax.get_transform("world"))


def test_iterate_coords(ignore_matplotlibrc):
    # Regression test for a bug that caused ax.coords to return too few axes

    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]
    wcs3d.wcs.cunit = ["deg", "deg", "km/s"]
    wcs3d.wcs.crpix = [614.5, 856.5, 333]
    wcs3d.wcs.cdelt = [6.25, 6.25, 23]
    wcs3d.wcs.crval = [0.0, 0.0, 1.0]

    ax = plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))

    x, y, z = ax.coords


def test_invalid_slices_errors(ignore_matplotlibrc):
    # Make sure that users get a clear message when specifying a WCS with
    # >2 dimensions without giving the 'slices' argument, or if the 'slices'
    # argument has too many/few elements.

    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]

    plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))

    with pytest.raises(
        ValueError,
        match=r"WCS has more than 2 pixel dimensions, so 'slices' should be set",
    ):
        plt.subplot(1, 1, 1, projection=wcs3d)

    with pytest.raises(
        ValueError,
        match=(
            r"'slices' should have as many elements as WCS has pixel dimensions .should"
            r" be 3."
        ),
    ):
        plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1, 2))

    wcs2d = WCS(naxis=2)
    wcs2d.wcs.ctype = ["x", "y"]

    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs2d)
    assert ax.frame_class is RectangularFrame
    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs2d, slices=("x", "y"))
    assert ax.frame_class is RectangularFrame
    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs2d, slices=("y", "x"))
    assert ax.frame_class is RectangularFrame
    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs2d, slices=["x", "y"])
    assert ax.frame_class is RectangularFrame
    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs2d, slices=(1, "x"))
    assert ax.frame_class is RectangularFrame1D

    wcs1d = WCS(naxis=1)
    wcs1d.wcs.ctype = ["x"]

    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs1d)
    assert ax.frame_class is RectangularFrame1D

    with pytest.raises(ValueError):
        plt.subplot(1, 1, 1, projection=wcs2d, slices=(1, "y"))


EXPECTED_REPR_1 = """
<CoordinatesMap with 3 world coordinates:

  index            aliases                type   ...    wrap   format_unit visible
  ----- ------------------------------ --------- ... --------- ----------- -------
      0                   distmod dist    scalar ...      None                  no
      1 pos.galactic.lon glon-car glon longitude ... 360.0 deg         deg     yes
      2 pos.galactic.lat glat-car glat  latitude ...      None         deg     yes

>
 """.strip()

EXPECTED_REPR_2 = """
<CoordinatesMap with 3 world coordinates:

  index            aliases                type   ...    wrap   format_unit visible
  ----- ------------------------------ --------- ... --------- ----------- -------
      0                   distmod dist    scalar ...      None                 yes
      1 pos.galactic.lon glon-car glon longitude ... 360.0 deg         deg     yes
      2 pos.galactic.lat glat-car glat  latitude ...      None         deg     yes

>
 """.strip()


def test_repr(ignore_matplotlibrc):
    # Unit test to make sure __repr__ looks as expected

    wcs3d = WCS(GAL_HEADER)

    # Cube header has world coordinates as distance, lon, lat, so start off
    # by slicing in a way that we select just lon,lat:

    ax = plt.subplot(1, 1, 1, projection=wcs3d, slices=(1, "x", "y"))
    assert repr(ax.coords) == EXPECTED_REPR_1

    # Now slice in a way that all world coordinates are still present:

    plt.clf()
    ax = plt.subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))
    assert repr(ax.coords) == EXPECTED_REPR_2


@pytest.fixture
def time_spectral_wcs_2d():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["FREQ", "TIME"]
    wcs.wcs.set()
    return wcs


def test_time_wcs(time_spectral_wcs_2d):
    # Regression test for a bug that caused WCSAxes to error when using a WCS
    # with a time axis.

    plt.subplot(projection=time_spectral_wcs_2d)


@pytest.mark.skipif(TEX_UNAVAILABLE, reason="TeX is unavailable")
def test_simplify_labels_usetex(ignore_matplotlibrc, tmp_path):
    """Regression test for https://github.com/astropy/astropy/issues/8004."""
    plt.rc("text", usetex=True)

    header = {
        "NAXIS": 2,
        "NAXIS1": 360,
        "NAXIS2": 180,
        "CRPIX1": 180.5,
        "CRPIX2": 90.5,
        "CRVAL1": 180.0,
        "CRVAL2": 0.0,
        "CDELT1": -2 * np.sqrt(2) / np.pi,
        "CDELT2": 2 * np.sqrt(2) / np.pi,
        "CTYPE1": "RA---MOL",
        "CTYPE2": "DEC--MOL",
        "RADESYS": "ICRS",
    }

    wcs = WCS(header)
    fig, ax = plt.subplots(subplot_kw=dict(frame_class=EllipticalFrame, projection=wcs))
    ax.set_xlim(-0.5, header["NAXIS1"] - 0.5)
    ax.set_ylim(-0.5, header["NAXIS2"] - 0.5)
    ax.coords[0].set_ticklabel(exclude_overlapping=True)
    ax.coords[1].set_ticklabel(exclude_overlapping=True)
    ax.coords[0].set_ticks(spacing=45 * u.deg)
    ax.coords[1].set_ticks(spacing=30 * u.deg)
    ax.grid()

    fig.savefig(tmp_path / "plot.png")


@pytest.mark.parametrize("frame_class", [RectangularFrame, EllipticalFrame])
def test_set_labels_with_coords(ignore_matplotlibrc, frame_class):
    """Test if ``axis.set_xlabel()`` calls the correct ``coords[i]_set_axislabel()`` in a
    WCS plot. Regression test for https://github.com/astropy/astropy/issues/10435.
    """

    labels = ["RA", "Declination"]
    header = {
        "NAXIS": 2,
        "NAXIS1": 360,
        "NAXIS2": 180,
        "CRPIX1": 180.5,
        "CRPIX2": 90.5,
        "CRVAL1": 180.0,
        "CRVAL2": 0.0,
        "CDELT1": -2 * np.sqrt(2) / np.pi,
        "CDELT2": 2 * np.sqrt(2) / np.pi,
        "CTYPE1": "RA---AIT",
        "CTYPE2": "DEC--AIT",
    }

    wcs = WCS(header)
    fig, ax = plt.subplots(subplot_kw=dict(frame_class=frame_class, projection=wcs))
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    assert ax.get_xlabel() == labels[0]
    assert ax.get_ylabel() == labels[1]
    for i in range(2):
        assert ax.coords[i].get_axislabel() == labels[i]


@pytest.mark.parametrize("atol", [0.2, 1.0e-8])
def test_bbox_size(atol):
    # Test for the size of a WCSAxes bbox (only have Matplotlib >= 3.0 now)
    extents = [11.38888888888889, 3.5, 576.0, 432.0]

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)
    fig.canvas.draw()
    renderer = fig.canvas.renderer
    ax_bbox = ax.get_tightbbox(renderer)

    # Enforce strict test only with reference Freetype version
    if atol < 0.1 and not FREETYPE_261:
        pytest.xfail(
            "Exact BoundingBox dimensions are only ensured with FreeType 2.6.1"
        )
    assert np.allclose(ax_bbox.extents, extents, atol=atol)


def test_wcs_type_transform_regression():
    wcs = WCS(TARGET_HEADER)
    sliced_wcs = SlicedLowLevelWCS(wcs, np.s_[1:-1, 1:-1])
    ax = plt.subplot(1, 1, 1, projection=wcs)
    ax.get_transform(sliced_wcs)

    high_wcs = HighLevelWCSWrapper(sliced_wcs)
    ax.get_transform(sliced_wcs)


def test_multiple_draws_grid_contours(tmp_path):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    ax.grid(color="black", grid_type="contours")
    fig.savefig(tmp_path / "plot.png")
    fig.savefig(tmp_path / "plot.png")


def test_get_coord_range_nan_regression():
    # Test to make sure there is no internal casting of NaN to integers
    # NumPy 1.24 raises a RuntimeWarning if a NaN is cast to an integer

    wcs = WCS(TARGET_HEADER)
    wcs.wcs.crval[0] = 0  # Re-position the longitude wrap to the middle
    ax = plt.subplot(1, 1, 1, projection=wcs)

    # Set the Y limits within valid latitudes/declinations
    ax.set_ylim(300, 500)

    # Set the X limits within valid longitudes/RAs, so the world coordinates have no NaNs
    ax.set_xlim(300, 700)
    assert np.allclose(
        ax.coords.get_coord_range(),
        np.array(
            [
                (-123.5219272110385, 122.49684897692201),
                (-44.02289164685554, 44.80732766607591),
            ]
        ),
    )

    # Extend the X limits to include invalid longitudes/RAs, so the world coordinates have NaNs
    ax.set_xlim(0, 700)
    assert np.allclose(
        ax.coords.get_coord_range(),
        np.array(
            [(-131.3193386797236, 180.0), (-44.02289164685554, 44.80732766607591)]
        ),
    )


def test_imshow_error():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    with pytest.raises(ValueError, match="Cannot use images with origin='upper"):
        ax.imshow(np.ones(100).reshape(10, 10), origin="upper")


def test_label_setting():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    # Check both xlabel and label kwargs work
    ax.set_xlabel(xlabel="label")
    ax.set_xlabel(label="label")
    # Check no label errors:
    with pytest.raises(
        TypeError, match=r"set_xlabel\(\) missing 1 required positional argument"
    ):
        ax.set_xlabel()

    # Check both xlabel and label kwargs work
    ax.set_ylabel(ylabel="label")
    ax.set_ylabel(label="label")
    # Check no label errors:
    with pytest.raises(
        TypeError, match=r"set_ylabel\(\) missing 1 required positional argument"
    ):
        ax.set_ylabel()


def test_invisible_bbox():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())

    assert ax.get_tightbbox(fig.canvas.get_renderer()) is not None
    ax.set_visible(False)
    assert ax.get_tightbbox(fig.canvas.get_renderer()) is None



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Tests for `position_angle()`, `separation()` and `separation_3d()` methods. They
are implemented in `BaseCoordinateFrame`, but are also exposed by `SkyCoord`
instances, so they should be tested on both.
"""

from contextlib import nullcontext
from typing import NamedTuple

import pytest

from astropy import units as u
from astropy.coordinates import (
    FK5,
    GCRS,
    ICRS,
    Angle,
    BaseCoordinateFrame,
    Distance,
    Galactic,
    NonRotationTransformationError,
    NonRotationTransformationWarning,
    SkyCoord,
)
from astropy.tests.helper import assert_quantity_allclose


class SeparationExpectation(NamedTuple):
    """
    The coordinates the position angle and separations are relative to
    are different for different tests.
    """

    coord: BaseCoordinateFrame | SkyCoord
    pytest_id: str
    position_angle: u.Quantity
    separation: u.Quantity
    separation_3d: u.Quantity
    reversed_position_angle: u.Quantity

    @property
    def reversed_separation(self) -> u.Quantity:
        return self.separation

    @property
    def reversed_separation_3d(self) -> u.Quantity:
        return self.separation_3d


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "other_coord",
    [
        SeparationExpectation(
            ICRS(0 * u.deg, 0 * u.deg, 3 * u.pc),
            "no_separation",
            0 * u.deg,
            0 * u.deg,
            0 * u.pc,
            0 * u.deg,
        ),
        SeparationExpectation(
            ICRS(0 * u.deg, 1 * u.deg, 3 * u.pc),
            "dec_separation",
            0 * u.deg,
            1 * u.deg,
            0.05235921 * u.pc,
            180 * u.deg,
        ),
        SeparationExpectation(
            ICRS(1 * u.deg, 0 * u.deg, 3 * u.pc),
            "ra_separation",
            90 * u.deg,
            1 * u.deg,
            0.05235921 * u.pc,
            270 * u.deg,
        ),
        SeparationExpectation(
            ICRS(0 * u.deg, 0 * u.deg, 10 * u.pc),
            "distance_separation",
            0 * u.deg,
            0 * u.deg,
            7 * u.pc,
            0 * u.deg,
        ),
        SeparationExpectation(
            SkyCoord(1 * u.deg, 1 * u.deg, 9 * u.pc),
            "SkyCoord_input",
            44.995636 * u.deg,
            1.4141777 * u.deg,
            6.00137 * u.pc,
            225.00436354 * u.deg,
        ),
    ],
    ids=lambda x: x.pytest_id,
)
@pytest.mark.parametrize("method", ["position_angle", "separation", "separation_3d"])
def test_scalar_coords(coord_class, other_coord, method):
    vernal_equinox = coord_class(0 * u.deg, 0 * u.deg, 3 * u.pc)
    assert_quantity_allclose(
        getattr(vernal_equinox, method)(other_coord.coord), getattr(other_coord, method)
    )
    assert_quantity_allclose(
        getattr(other_coord.coord, method)(vernal_equinox),
        getattr(other_coord, f"reversed_{method}"),
    )


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "other_coord",
    [
        SeparationExpectation(
            FK5(0 * u.deg, 90 * u.deg, 2 * u.pc),
            "FK5_input",
            245.42603114 * u.deg,
            6.07832112e-06 * u.deg,
            2.12173433e-07 * u.pc,
            65.42602474 * u.deg,
        ),
        SeparationExpectation(
            Galactic(0 * u.deg, 90 * u.deg, 9 * u.pc),
            "Galactic_input",
            347.14052211 * u.deg,
            62.871748 * u.deg,
            8.28158093 * u.pc,
            57.06807474 * u.deg,
        ),
    ],
    ids=lambda x: x.pytest_id,
)
@pytest.mark.parametrize("method", ["position_angle", "separation", "separation_3d"])
def test_scalar_coords_frame_transformation(coord_class, other_coord, method):
    north_pole = coord_class(0 * u.deg, 90 * u.deg, 2 * u.pc)
    assert_quantity_allclose(
        getattr(north_pole, method)(other_coord.coord), getattr(other_coord, method)
    )
    assert_quantity_allclose(
        getattr(other_coord.coord, method)(north_pole),
        getattr(other_coord, f"reversed_{method}"),
    )


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "other_coord",
    [
        SeparationExpectation(
            ICRS([-1, -2, -3] * u.deg, [0.1, 1.1, 2.1] * u.deg, [11, 13, 17] * u.pc),
            "ICRS_input",
            [275.710887, 272.880921, 271.963607] * u.deg,
            [1.0049871, 2.0021627, 2.9997464] * u.deg,
            [8.000635, 8.004959, 10.016293] * u.pc,
            [95.71001423, 92.84426777, 91.8562681] * u.deg,
        ),
        SeparationExpectation(
            SkyCoord([1, -2, 3] * u.deg, [-0.1, 1.1, 2.1] * u.deg, [11, 13, 17] * u.pc),
            "SkyCoord_input",
            [95.71088692, 272.880921, 88.03639251] * u.deg,
            [1.0049871, 2.0021627, 2.9997464] * u.deg,
            [8.000635, 8.004959, 10.016293] * u.pc,
            [275.71001423, 92.84426777, 268.1437319] * u.deg,
        ),
    ],
    ids=lambda x: x.pytest_id,
)
@pytest.mark.parametrize("method", ["position_angle", "separation", "separation_3d"])
def test_array_coords(coord_class, other_coord, method):
    coord = coord_class(0 * u.deg, [0, 1, 2] * u.deg, [3, 5, 7] * u.pc)
    assert_quantity_allclose(
        getattr(coord, method)(other_coord.coord), getattr(other_coord, method)
    )
    assert_quantity_allclose(
        getattr(other_coord.coord, method)(coord),
        getattr(other_coord, f"reversed_{method}"),
    )


@pytest.mark.parametrize(
    "coord",
    [
        pytest.param(FK5(1 * u.deg, 0 * u.deg, 1 * u.pc), id="FK5"),
        pytest.param(
            SkyCoord(1 * u.deg, 0 * u.deg, 1 * u.pc, frame="fk5"), id="SkyCoord"
        ),
    ],
)
@pytest.mark.parametrize(
    "other_coord",
    [
        SeparationExpectation(
            FK5(1 * u.deg, 0 * u.deg, 1 * u.pc, equinox="B1950"),
            "FK5_B1950",
            66.51310007 * u.deg,
            0.69835342 * u.deg,
            0.01218849 * u.pc,
            246.50823798 * u.deg,
        ),
        SeparationExpectation(
            SkyCoord(1 * u.deg, 0 * u.deg, 10 * u.pc, frame="fk5", equinox="B1950"),
            "SkyCoord_B1950",
            66.51310007 * u.deg,
            0.69835342 * u.deg,
            9.000083 * u.pc,
            246.50823798 * u.deg,
        ),
    ],
    ids=lambda x: x.pytest_id,
)
@pytest.mark.parametrize("method", ["position_angle", "separation", "separation_3d"])
def test_equinox_conversion(coord, other_coord, method):
    """
    Regression test for
        - #868, #891: methods raised errors in case of frames with equinoxes
        - #3106, #15659: unspecified equinox should be interpreted as frame default,
          both in `SkyCoord` methods (#3106) and in `BaseCoordinateFrame` methods with
          `SkyCoord` input (#15659).
        - #5702: `position_angle()` didn't check if equinoxes differed
    """
    assert_quantity_allclose(
        getattr(coord, method)(other_coord.coord), getattr(other_coord, method)
    )
    assert_quantity_allclose(
        getattr(other_coord.coord, method)(coord),
        getattr(other_coord, f"reversed_{method}"),
    )


@pytest.mark.parametrize("other_class", [SkyCoord, ICRS])
@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
def test_separation_3d_dimensionless_distance(coord_class, other_class):
    assert_quantity_allclose(
        coord_class(35 * u.deg, 0 * u.deg, 3 * u.one).separation_3d(
            other_class(125 * u.deg, 0 * u.deg, 4 * u.one)
        ),
        5 * u.one,
    )


@pytest.mark.parametrize("dimensionless_class", [SkyCoord, ICRS])
@pytest.mark.parametrize("length_class", [SkyCoord, ICRS])
def test_separation_3d_distance_dimension_mismatch(length_class, dimensionless_class):
    dimensionless_coord = dimensionless_class(1 * u.deg, -2 * u.deg, 14)
    length_coord = length_class(-1 * u.deg, 2 * u.deg, 21 * u.pc)
    error_message = (
        "^Can only apply 'subtract' function to quantities with compatible dimensions$"
    )
    with pytest.raises(u.UnitConversionError, match=error_message):
        dimensionless_coord.separation_3d(length_coord)
    with pytest.raises(u.UnitConversionError, match=error_message):
        length_coord.separation_3d(dimensionless_coord)


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
def test_separation_3d_no_distance(coord_class):
    coord_no_distance = coord_class(0 * u.deg, 0 * u.deg)
    coord_with_distance = ICRS(0 * u.deg, 0 * u.deg, distance=3 * u.pc)
    with pytest.raises(
        ValueError,
        match="^This object does not have a distance; cannot compute 3d separation.$",
    ):
        coord_no_distance.separation_3d(coord_with_distance)
    with pytest.raises(
        u.UnitConversionError,
        match=(
            "^Can only apply 'subtract' function to quantities with compatible"
            " dimensions$"
        ),
    ):
        coord_with_distance.separation_3d(coord_no_distance)


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "velocity_kwargs",
    [
        pytest.param({}, id="no_velocity"),
        pytest.param({"radial_velocity": -108 * u.km / u.s}, id="radial_velocity"),
        pytest.param(
            {"pm_ra_cosdec": 7 * u.mas / u.s, "pm_dec": -5 * u.mas / u.s},
            id="proper_motion",
        ),
        pytest.param(
            {
                "radial_velocity": -108 * u.km / u.s,
                "pm_ra_cosdec": 7 * u.mas / u.s,
                "pm_dec": -5 * u.mas / u.s,
            },
            id="3d_velocity",
        ),
    ],
)
@pytest.mark.parametrize(
    "other_coord",
    [
        SeparationExpectation(
            ICRS(
                ra=20 * u.deg,
                dec=10 * u.deg,
                distance=8 * u.pc,
                pm_ra_cosdec=14 * u.mas / u.yr,
                pm_dec=-11 * u.mas / u.yr,
                radial_velocity=5 * u.km / u.s,
            ),
            "ICRS_with_velocity",
            134.58168775 * u.deg,
            13.89233851 * u.deg,
            3.36750833 * u.pc,
            317.18593154 * u.deg,
        ),
        SeparationExpectation(
            SkyCoord(ra=20 * u.deg, dec=10 * u.deg, distance=8 * u.pc),
            "SkyCoord_no_velocity",
            134.58168775 * u.deg,
            13.89233851 * u.deg,
            3.36750833 * u.pc,
            317.18593154 * u.deg,
        ),
    ],
    ids=lambda x: x.pytest_id,
)
@pytest.mark.parametrize("method", ["position_angle", "separation", "separation_3d"])
def test_with_velocities(coord_class, velocity_kwargs, other_coord, method):
    coord = coord_class(
        ra=10 * u.deg, dec=20 * u.deg, distance=5 * u.pc, **velocity_kwargs
    )
    assert_quantity_allclose(
        getattr(coord, method)(other_coord.coord), getattr(other_coord, method)
    )
    assert_quantity_allclose(
        getattr(other_coord.coord, method)(coord),
        getattr(other_coord, f"reversed_{method}"),
    )


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "method,output_type",
    [
        pytest.param(method, type_, id=method)
        for method, type_ in (
            ("position_angle", Angle),
            ("separation", Angle),
            ("separation_3d", Distance),
        )
    ],
)
def test_return_types(coord_class, method, output_type):
    """
    This is especially important for SkyCoord because SkyCoord instances
    expose the methods of their underlying frame at runtime, so they cannot be
    checked statically.
    """
    coord = coord_class(0 * u.deg, 0 * u.deg, 1 * u.pc)
    assert type(getattr(coord, method)(coord)) is output_type


@pytest.mark.parametrize("coord_class", [SkyCoord, ICRS])
@pytest.mark.parametrize(
    "origin_mismatch_kwarg,expectation",
    [
        pytest.param({"origin_mismatch": "ignore"}, nullcontext(), id="ignore"),
        pytest.param(
            {"origin_mismatch": "warn"},
            pytest.warns(
                NonRotationTransformationWarning,
                match="^transforming other coordinates from <GCRS Frame ",
            ),
            id="warn",
        ),
        pytest.param(
            {"origin_mismatch": "error"},
            pytest.raises(
                NonRotationTransformationError,
                match="^refusing to transform other coordinates from <GCRS Frame ",
            ),
            id="error",
        ),
        pytest.param(
            {},
            pytest.warns(
                NonRotationTransformationWarning,
                match="^transforming other coordinates from <GCRS Frame ",
            ),
            id="default",
        ),
        pytest.param(
            {"origin_mismatch": "bad"},
            pytest.raises(
                ValueError,
                match=(
                    r"^origin_mismatch='bad' is invalid\. Allowed values are 'ignore', "
                    r"'warn' or 'error'\.$"
                ),
            ),
            id="invalid",
        ),
    ],
)
def test_separation_origin_mismatch_action(
    coord_class, origin_mismatch_kwarg, expectation
):
    with expectation:
        coord_class(0 * u.deg, 0 * u.deg).separation(
            SkyCoord(0 * u.deg, 0 * u.deg, frame=GCRS), **origin_mismatch_kwarg
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import socketserver
import sys
import traceback
import warnings
import xmlrpc.client as xmlrpc
from xmlrpc.server import SimpleXMLRPCRequestHandler, SimpleXMLRPCServer

from .constants import SAMP_ICON
from .errors import SAMPWarning

__all__ = []


class SAMPSimpleXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
    """
    XMLRPC handler of Standard Profile requests.
    """

    def do_GET(self):
        if self.path == "/samp/icon":
            self.send_response(200, "OK")
            self.send_header("Content-Type", "image/png")
            self.end_headers()
            self.wfile.write(SAMP_ICON)

    def do_POST(self):
        """
        Handles the HTTP POST request.

        Attempts to interpret all HTTP POST requests as XML-RPC calls,
        which are forwarded to the server's ``_dispatch`` method for
        handling.
        """
        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.report_404()
            return

        try:
            # Get arguments by reading body of request.
            # We read this in chunks to avoid straining
            # socket.read(); around the 10 or 15Mb mark, some platforms
            # begin to have problems (bug #792570).
            max_chunk_size = 10 * 1024 * 1024
            size_remaining = int(self.headers["content-length"])
            L = []
            while size_remaining:
                chunk_size = min(size_remaining, max_chunk_size)
                L.append(self.rfile.read(chunk_size))
                size_remaining -= len(L[-1])
            data = b"".join(L)

            params, method = xmlrpc.loads(data)

            if method == "samp.webhub.register":
                params = list(params)
                params.append(self.client_address)
                if "Origin" in self.headers:
                    params.append(self.headers.get("Origin"))
                else:
                    params.append("unknown")
                params = tuple(params)
                data = xmlrpc.dumps(params, methodname=method)

            elif method in (
                "samp.hub.notify",
                "samp.hub.notifyAll",
                "samp.hub.call",
                "samp.hub.callAll",
                "samp.hub.callAndWait",
            ):
                user = "unknown"

                if method == "samp.hub.callAndWait":
                    params[2]["host"] = self.address_string()
                    params[2]["user"] = user
                else:
                    params[-1]["host"] = self.address_string()
                    params[-1]["user"] = user

                data = xmlrpc.dumps(params, methodname=method)

            data = self.decode_request_content(data)
            if data is None:
                return  # response has been sent

            # In previous versions of SimpleXMLRPCServer, _dispatch
            # could be overridden in this class, instead of in
            # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
            # check to see if a subclass implements _dispatch and dispatch
            # using that method if present.
            response = self.server._marshaled_dispatch(
                data, getattr(self, "_dispatch", None), self.path
            )
        except Exception as e:
            # This should only happen if the module is buggy
            # internal error, report as HTTP server error
            self.send_response(500)

            # Send information about the exception if requested
            if (
                hasattr(self.server, "_send_traceback_header")
                and self.server._send_traceback_header
            ):
                self.send_header("X-exception", str(e))
                trace = traceback.format_exc()
                trace = str(trace.encode("ASCII", "backslashreplace"), "ASCII")
                self.send_header("X-traceback", trace)

            self.send_header("Content-length", "0")
            self.end_headers()
        else:
            # got a valid XML RPC response
            self.send_response(200)
            self.send_header("Content-type", "text/xml")
            if self.encode_threshold is not None:
                if len(response) > self.encode_threshold:
                    q = self.accept_encodings().get("gzip", 0)
                    if q:
                        try:
                            response = xmlrpc.gzip_encode(response)
                            self.send_header("Content-Encoding", "gzip")
                        except NotImplementedError:
                            pass
            self.send_header("Content-length", str(len(response)))
            self.end_headers()
            self.wfile.write(response)


class ThreadingXMLRPCServer(socketserver.ThreadingMixIn, SimpleXMLRPCServer):
    """
    Asynchronous multithreaded XMLRPC server.
    """

    def __init__(
        self,
        addr,
        log=None,
        requestHandler=SAMPSimpleXMLRPCRequestHandler,
        logRequests=True,
        allow_none=True,
        encoding=None,
    ):
        self.log = log
        SimpleXMLRPCServer.__init__(
            self, addr, requestHandler, logRequests, allow_none, encoding
        )

    def handle_error(self, request, client_address):
        if self.log is None:
            socketserver.BaseServer.handle_error(self, request, client_address)
        else:
            warnings.warn(
                (
                    "Exception happened during processing of request from "
                    f"{client_address}: {sys.exc_info()[1]}"
                ),
                SAMPWarning,
            )



"""Testing :mod:`astropy.cosmology.units`."""

import pytest

import astropy.cosmology.units as cu
import astropy.units as u
from astropy.cosmology import Planck13, default_cosmology
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY

##############################################################################
# TESTS
##############################################################################


def test_littleh():
    """Test :func:`astropy.cosmology.units.with_H0`."""
    H0_70 = 70 * u.km / u.s / u.Mpc
    h70dist = 70 * u.Mpc / cu.littleh

    assert_quantity_allclose(h70dist.to(u.Mpc, cu.with_H0(H0_70)), 100 * u.Mpc)

    # make sure using the default cosmology works
    cosmodist = default_cosmology.get().H0.value * u.Mpc / cu.littleh
    assert_quantity_allclose(cosmodist.to(u.Mpc, cu.with_H0()), 100 * u.Mpc)

    # Now try a luminosity scaling
    h1lum = 0.49 * u.Lsun * cu.littleh**-2
    assert_quantity_allclose(h1lum.to(u.Lsun, cu.with_H0(H0_70)), 1 * u.Lsun)

    # And the trickiest one: magnitudes.  Using H0=10 here for the round numbers
    H0_10 = 10 * u.km / u.s / u.Mpc
    # assume the "true" magnitude M = 12.
    # Then M - 5*log_10(h)  = M + 5 = 17
    withlittlehmag = 17 * (u.mag - u.MagUnit(cu.littleh**2))
    assert_quantity_allclose(withlittlehmag.to(u.mag, cu.with_H0(H0_10)), 12 * u.mag)


@pytest.mark.skipif(not HAS_SCIPY, reason="Cosmology needs scipy")
def test_dimensionless_redshift():
    """Test :func:`astropy.cosmology.units.dimensionless_redshift`."""
    z = 3 * cu.redshift
    val = 3 * u.one

    # show units not equal
    assert z.unit == cu.redshift
    assert z.unit != u.one
    assert u.get_physical_type(z) == "redshift"

    # test equivalency enabled by default
    assert z == val

    # also test that it works for powers
    assert (3 * cu.redshift**3) == val

    # and in composite units
    assert (3 * u.km / cu.redshift**3) == 3 * u.km

    # test it also works as an equivalency
    with u.set_enabled_equivalencies([]):  # turn off default equivalencies
        assert z.to(u.one, equivalencies=cu.dimensionless_redshift()) == val

        with pytest.raises(ValueError):
            z.to(u.one)

    # if this fails, something is really wrong
    with u.add_enabled_equivalencies(cu.dimensionless_redshift()):
        assert z == val


@pytest.mark.skipif(not HAS_SCIPY, reason="Cosmology needs scipy")
def test_redshift_temperature():
    """Test :func:`astropy.cosmology.units.redshift_temperature`."""
    cosmo = Planck13.clone(Tcmb0=3 * u.K)
    default_cosmo = default_cosmology.get()
    z = 15 * cu.redshift
    Tcmb = cosmo.Tcmb(z)

    # 1) Default (without specifying the cosmology)
    with default_cosmology.set(cosmo):
        equivalency = cu.redshift_temperature()
        assert_quantity_allclose(z.to(u.K, equivalency), Tcmb)
        assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)

    # showing the answer changes if the cosmology changes
    # this test uses the default cosmology
    equivalency = cu.redshift_temperature()
    assert_quantity_allclose(z.to(u.K, equivalency), default_cosmo.Tcmb(z))
    assert default_cosmo.Tcmb(z) != Tcmb

    # 2) Specifying the cosmology
    equivalency = cu.redshift_temperature(cosmo)
    assert_quantity_allclose(z.to(u.K, equivalency), Tcmb)
    assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)

    # Test `atzkw`
    equivalency = cu.redshift_temperature(cosmo, ztol=1e-10)
    assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)


@pytest.mark.skipif(not HAS_SCIPY, reason="Cosmology needs scipy")
def test_redshift_hubble():
    """Test :func:`astropy.cosmology.units.redshift_hubble`."""
    unit = u.km / u.s / u.Mpc
    cosmo = Planck13.clone(H0=100 * unit)
    default_cosmo = default_cosmology.get()
    z = 15 * cu.redshift
    H = cosmo.H(z)
    h = H.to_value(u.km / u.s / u.Mpc) / 100 * cu.littleh

    # 1) Default (without specifying the cosmology)
    with default_cosmology.set(cosmo):
        equivalency = cu.redshift_hubble()
        # H
        assert_quantity_allclose(z.to(unit, equivalency), H)
        assert_quantity_allclose(H.to(cu.redshift, equivalency), z)
        # little-h
        assert_quantity_allclose(z.to(cu.littleh, equivalency), h)
        assert_quantity_allclose(h.to(cu.redshift, equivalency), z)

    # showing the answer changes if the cosmology changes
    # this test uses the default cosmology
    equivalency = cu.redshift_hubble()
    assert_quantity_allclose(z.to(unit, equivalency), default_cosmo.H(z))
    assert default_cosmo.H(z) != H

    # 2) Specifying the cosmology
    equivalency = cu.redshift_hubble(cosmo)
    # H
    assert_quantity_allclose(z.to(unit, equivalency), H)
    assert_quantity_allclose(H.to(cu.redshift, equivalency), z)
    # little-h
    assert_quantity_allclose(z.to(cu.littleh, equivalency), h)
    assert_quantity_allclose(h.to(cu.redshift, equivalency), z)

    # Test `atzkw`
    equivalency = cu.redshift_hubble(cosmo, ztol=1e-10)
    assert_quantity_allclose(H.to(cu.redshift, equivalency), z)  # H
    assert_quantity_allclose(h.to(cu.redshift, equivalency), z)  # little-h


@pytest.mark.skipif(not HAS_SCIPY, reason="Cosmology needs scipy")
@pytest.mark.parametrize(
    "kind",
    [cu.redshift_distance.__defaults__[-1], "comoving", "lookback", "luminosity"],
)
def test_redshift_distance(kind):
    """Test :func:`astropy.cosmology.units.redshift_distance`."""
    z = 15 * cu.redshift
    d = getattr(Planck13, kind + "_distance")(z)

    equivalency = cu.redshift_distance(cosmology=Planck13, kind=kind)

    # properties of Equivalency
    assert equivalency.name[0] == "redshift_distance"
    assert equivalency.kwargs[0]["cosmology"] == Planck13
    assert equivalency.kwargs[0]["distance"] == kind

    # roundtrip
    assert_quantity_allclose(z.to(u.Mpc, equivalency), d)
    assert_quantity_allclose(d.to(cu.redshift, equivalency), z)


def test_redshift_distance_wrong_kind():
    """Test :func:`astropy.cosmology.units.redshift_distance` wrong kind."""
    with pytest.raises(ValueError, match="`kind`"):
        cu.redshift_distance(kind=None)


@pytest.mark.skipif(not HAS_SCIPY, reason="Cosmology needs scipy")
class Test_with_redshift:
    """Test `astropy.cosmology.units.with_redshift`."""

    @pytest.fixture(scope="class")
    def cosmo(self):
        """Test cosmology."""
        return Planck13.clone(Tcmb0=3 * u.K)

    # ===========================================

    def test_cosmo_different(self, cosmo):
        """The default is different than the test cosmology."""
        default_cosmo = default_cosmology.get()
        assert default_cosmo != cosmo  # shows changing default

    def test_no_equivalency(self, cosmo):
        """Test the equivalency  ``with_redshift`` without any enabled."""
        equivalency = cu.with_redshift(distance=None, hubble=False, Tcmb=False)
        assert len(equivalency) == 0

    # -------------------------------------------

    def test_temperature_off(self, cosmo):
        """Test ``with_redshift`` with the temperature off."""
        z = 15 * cu.redshift
        err_msg = (
            r"^'redshift' \(redshift\) and 'K' \(temperature\) are not convertible$"
        )

        # 1) Default (without specifying the cosmology)
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(Tcmb=False)
            with pytest.raises(u.UnitConversionError, match=err_msg):
                z.to(u.K, equivalency)

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, Tcmb=False)
        with pytest.raises(u.UnitConversionError, match=err_msg):
            z.to(u.K, equivalency)

    def test_temperature(self, cosmo):
        """Test temperature equivalency component."""
        default_cosmo = default_cosmology.get()
        z = 15 * cu.redshift
        Tcmb = cosmo.Tcmb(z)

        # 1) Default (without specifying the cosmology)
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(Tcmb=True)
            assert_quantity_allclose(z.to(u.K, equivalency), Tcmb)
            assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)

        # showing the answer changes if the cosmology changes
        # this test uses the default cosmology
        equivalency = cu.with_redshift(Tcmb=True)
        assert_quantity_allclose(z.to(u.K, equivalency), default_cosmo.Tcmb(z))
        assert default_cosmo.Tcmb(z) != Tcmb

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, Tcmb=True)
        assert_quantity_allclose(z.to(u.K, equivalency), Tcmb)
        assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)

        # Test `atzkw`
        # this is really just a test that 'atzkw' doesn't fail
        equivalency = cu.with_redshift(cosmo, Tcmb=True, atzkw={"ztol": 1e-10})
        assert_quantity_allclose(Tcmb.to(cu.redshift, equivalency), z)

    # -------------------------------------------

    def test_hubble_off(self, cosmo):
        """Test ``with_redshift`` with Hubble off."""
        unit = u.km / u.s / u.Mpc
        z = 15 * cu.redshift
        err_msg = (
            r"^'redshift' \(redshift\) and 'km / \(Mpc s\)' \(frequency\) are not "
            "convertible$"
        )

        # 1) Default (without specifying the cosmology)
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(hubble=False)
            with pytest.raises(u.UnitConversionError, match=err_msg):
                z.to(unit, equivalency)

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, hubble=False)
        with pytest.raises(u.UnitConversionError, match=err_msg):
            z.to(unit, equivalency)

    def test_hubble(self, cosmo):
        """Test Hubble equivalency component."""
        unit = u.km / u.s / u.Mpc
        default_cosmo = default_cosmology.get()
        z = 15 * cu.redshift
        H = cosmo.H(z)
        h = H.to_value(u.km / u.s / u.Mpc) / 100 * cu.littleh

        # 1) Default (without specifying the cosmology)
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(hubble=True)
            # H
            assert_quantity_allclose(z.to(unit, equivalency), H)
            assert_quantity_allclose(H.to(cu.redshift, equivalency), z)
            # little-h
            assert_quantity_allclose(z.to(cu.littleh, equivalency), h)
            assert_quantity_allclose(h.to(cu.redshift, equivalency), z)

        # showing the answer changes if the cosmology changes
        # this test uses the default cosmology
        equivalency = cu.with_redshift(hubble=True)
        assert_quantity_allclose(z.to(unit, equivalency), default_cosmo.H(z))
        assert default_cosmo.H(z) != H

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, hubble=True)
        # H
        assert_quantity_allclose(z.to(unit, equivalency), H)
        assert_quantity_allclose(H.to(cu.redshift, equivalency), z)
        # little-h
        assert_quantity_allclose(z.to(cu.littleh, equivalency), h)
        assert_quantity_allclose(h.to(cu.redshift, equivalency), z)

        # Test `atzkw`
        # this is really just a test that 'atzkw' doesn't fail
        equivalency = cu.with_redshift(cosmo, hubble=True, atzkw={"ztol": 1e-10})
        assert_quantity_allclose(H.to(cu.redshift, equivalency), z)  # H
        assert_quantity_allclose(h.to(cu.redshift, equivalency), z)  # h

    # -------------------------------------------

    def test_distance_off(self, cosmo):
        """Test ``with_redshift`` with the distance off."""
        z = 15 * cu.redshift
        err_msg = r"^'redshift' \(redshift\) and 'Mpc' \(length\) are not convertible$"

        # 1) Default (without specifying the cosmology)
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(distance=None)
            with pytest.raises(u.UnitConversionError, match=err_msg):
                z.to(u.Mpc, equivalency)

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, distance=None)
        with pytest.raises(u.UnitConversionError, match=err_msg):
            z.to(u.Mpc, equivalency)

    def test_distance_default(self):
        """Test distance equivalency default."""
        z = 15 * cu.redshift
        d = default_cosmology.get().comoving_distance(z)

        equivalency = cu.with_redshift()
        assert_quantity_allclose(z.to(u.Mpc, equivalency), d)
        assert_quantity_allclose(d.to(cu.redshift, equivalency), z)

    def test_distance_wrong_kind(self):
        """Test distance equivalency, but the wrong kind."""
        with pytest.raises(ValueError, match="`kind`"):
            cu.with_redshift(distance=ValueError)

    @pytest.mark.parametrize("kind", ["comoving", "lookback", "luminosity"])
    def test_distance(self, kind):
        """Test distance equivalency."""
        cosmo = Planck13
        z = 15 * cu.redshift
        dist = getattr(cosmo, kind + "_distance")(z)

        default_cosmo = default_cosmology.get()
        assert default_cosmo != cosmo  # shows changing default

        # 1) without specifying the cosmology
        with default_cosmology.set(cosmo):
            equivalency = cu.with_redshift(distance=kind)
            assert_quantity_allclose(z.to(u.Mpc, equivalency), dist)

        # showing the answer changes if the cosmology changes
        # this test uses the default cosmology
        equivalency = cu.with_redshift(distance=kind)
        assert_quantity_allclose(
            z.to(u.Mpc, equivalency), getattr(default_cosmo, kind + "_distance")(z)
        )
        assert not u.allclose(getattr(default_cosmo, kind + "_distance")(z), dist)

        # 2) Specifying the cosmology
        equivalency = cu.with_redshift(cosmo, distance=kind)
        assert_quantity_allclose(z.to(u.Mpc, equivalency), dist)
        assert_quantity_allclose(dist.to(cu.redshift, equivalency), z)

        # Test atzkw
        # this is really just a test that 'atzkw' doesn't fail
        equivalency = cu.with_redshift(cosmo, distance=kind, atzkw={"ztol": 1e-10})
        assert_quantity_allclose(dist.to(cu.redshift, equivalency), z)


def test_equivalency_context_manager():
    base_registry = u.get_current_unit_registry()

    # check starting with only the dimensionless_redshift equivalency.
    assert len(base_registry.equivalencies) == 1
    assert str(base_registry.equivalencies[0][0]) == "redshift"



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from contextlib import contextmanager
from functools import wraps
from types import FunctionType

from astropy.table import QTable

__all__ = ["BaseTimeSeries", "autocheck_required_columns"]

COLUMN_RELATED_METHODS = [
    "add_column",
    "add_columns",
    "keep_columns",
    "remove_column",
    "remove_columns",
    "rename_column",
]


def autocheck_required_columns(cls):
    """
    This is a decorator that ensures that the table contains specific
    methods indicated by the _required_columns attribute. The aim is to
    decorate all methods that might affect the columns in the table and check
    for consistency after the methods have been run.
    """

    def decorator_method(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            result = method(self, *args, **kwargs)
            self._check_required_columns()
            return result

        return wrapper

    for name in COLUMN_RELATED_METHODS:
        if not hasattr(cls, name) or not isinstance(getattr(cls, name), FunctionType):
            raise ValueError(f"{name} is not a valid method")
        setattr(cls, name, decorator_method(getattr(cls, name)))

    return cls


class BaseTimeSeries(QTable):
    _required_columns = None
    _required_columns_enabled = True

    # If _required_column_relax is True, we don't require the columns to be
    # present but we do require them to be the correct ones IF present. Note
    # that this is a temporary state - as soon as the required columns
    # are all present, we toggle this to False
    _required_columns_relax = False

    def _check_required_columns(self):
        def as_scalar_or_list_str(obj):
            if not hasattr(obj, "__len__"):
                return f"'{obj}'"
            elif len(obj) == 1:
                return f"'{obj[0]}'"
            else:
                return str(obj)

        if not self._required_columns_enabled:
            return

        if self._required_columns is not None:
            if self._required_columns_relax:
                required_columns = self._required_columns[: len(self.colnames)]
            else:
                required_columns = self._required_columns

            plural = "s" if len(required_columns) > 1 else ""

            if not self._required_columns_relax and len(self.colnames) == 0:
                raise ValueError(
                    f"{self.__class__.__name__} object is invalid - expected"
                    f" '{required_columns[0]}' as the first column{plural} but time"
                    " series has no columns"
                )

            elif self.colnames[: len(required_columns)] != required_columns:
                raise ValueError(
                    f"{self.__class__.__name__} object is invalid - expected"
                    f" {as_scalar_or_list_str(required_columns)} as the first"
                    f" column{plural} but found"
                    f" {as_scalar_or_list_str(self.colnames[: len(required_columns)])}"
                )

            if (
                self._required_columns_relax
                and self._required_columns
                == self.colnames[: len(self._required_columns)]
            ):
                self._required_columns_relax = False

    @contextmanager
    def _delay_required_column_checks(self):
        self._required_columns_enabled = False
        yield
        self._required_columns_enabled = True
        self._check_required_columns()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package defines SI prefixed units that are required by the VOUnit standard
but that are rarely used in practice and liable to lead to confusion (such as
``msolMass`` for milli-solar mass). They are in a separate module from
`astropy.units.deprecated` because they need to be enabled by default for
`astropy.units` to parse compliant VOUnit strings. As a result, e.g.,
``Unit('msolMass')`` will just work, but to access the unit directly, use
``astropy.units.required_by_vounit.msolMass`` instead of the more typical idiom
possible for the non-prefixed unit, ``astropy.units.solMass``.
"""

_ns = globals()


def _initialize_module():
    # Local imports to avoid polluting top-level namespace
    from . import astrophys
    from .core import _add_prefixes

    _add_prefixes(astrophys.solMass, namespace=_ns, prefixes=True)
    _add_prefixes(astrophys.solRad, namespace=_ns, prefixes=True)
    _add_prefixes(astrophys.solLum, namespace=_ns, prefixes=True)


_initialize_module()


###########################################################################
# DOCSTRING

if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    from .utils import (
        generate_prefixonly_unit_summary as _generate_prefixonly_unit_summary,
    )
    from .utils import generate_unit_summary as _generate_unit_summary

    __doc__ += _generate_unit_summary(globals())
    __doc__ += _generate_prefixonly_unit_summary(globals())


def _enable():
    """
    Enable the VOUnit-required extra units so they appear in results of
    `~astropy.units.UnitBase.find_equivalent_units` and
    `~astropy.units.UnitBase.compose`, and are recognized in the ``Unit('...')``
    idiom.
    """
    # Local import to avoid cyclical import
    # Local import to avoid polluting namespace
    import inspect

    from .core import add_enabled_units

    return add_enabled_units(inspect.getmodule(_enable))


# Because these are VOUnit mandated units, they start enabled (which is why the
# function is hidden).
_enable()



import hashlib
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy.io import fits
from astropy.io.fits.hdu.compressed._codecs import PLIO1
from astropy.io.fits.hdu.compressed._compression import CfitsioException

from .conftest import fitsio_param_to_astropy_param


@pytest.fixture
def canonical_data_base_path():
    return Path(__file__).parent / "data"


@pytest.fixture(
    params=(Path(__file__).parent / "data").glob("m13_*.fits"), ids=lambda x: x.name
)
def canonical_int_hdus(request):
    """
    This fixture provides 4 files downloaded from https://fits.gsfc.nasa.gov/registry/tilecompression.html

    Which are used as canonical tests of data not compressed by Astropy.
    """
    with fits.open(request.param) as hdul:
        yield hdul[1]


@pytest.fixture
def original_int_hdu(canonical_data_base_path):
    with fits.open(canonical_data_base_path / "m13.fits") as hdul:
        yield hdul[0]


def test_canonical_data(original_int_hdu, canonical_int_hdus):
    assert_allclose(original_int_hdu.data, canonical_int_hdus.data)


def test_zblank_support(canonical_data_base_path, tmp_path):
    # This uses a test 12x12 image which contains a NaN value in the [1, 1]
    # pixel - it was compressed using fpack which automatically added a ZBLANK
    # header keyword

    reference = np.arange(144).reshape((12, 12)).astype(float)
    reference[1, 1] = np.nan

    with fits.open(canonical_data_base_path / "compressed_with_nan.fits") as hdul:
        assert_equal(np.round(hdul[1].data), reference)

    # Now generate a file ourselves and check that the output has the ZBLANK
    # keyword set automatically

    hdu = fits.CompImageHDU(
        data=reference, compression_type="RICE_1", tile_shape=(6, 6)
    )

    hdu.writeto(tmp_path / "test_zblank.fits")

    with fits.open(tmp_path / "test_zblank.fits") as hdul:
        assert "ZBLANK" in hdul[1].header
        assert_equal(np.round(hdul[1].data), reference)


@pytest.mark.parametrize(
    ("shape", "tile_shape"),
    (
        ([10, 10], [5, 5]),  # something for HCOMPRESS
        ([5, 5, 5], [5, 5, 5]),
        # ([5, 5, 5], [5, 5, 1]),  # something for HCOMPRESS
        ([10, 15, 20], [5, 5, 5]),
        ([10, 5, 12], [5, 5, 5]),
        # TODO: There's a stupid bit of code in CompImageHDU which stops this working.
        # ([2, 3, 4, 5], [1, 1, 2, 3]),
        ([2, 3, 4, 5], [5, 5, 1, 1]),
    ),
)
def test_roundtrip_high_D(
    numpy_rng, compression_type, compression_param, tmp_path, dtype, shape, tile_shape
):
    if compression_type == "HCOMPRESS_1" and (
        # We don't have at least a 2D image
        len(shape) < 2
        or
        # We don't have 2D tiles
        np.count_nonzero(np.array(tile_shape) != 1) != 2
        or
        # TODO: The following restrictions can be lifted with some extra work.
        # The tile is not the first two dimensions of the data
        tile_shape[0] == 1
        or tile_shape[1] == 1
        or
        # The tile dimensions not an integer multiple of the array dims
        np.count_nonzero(np.array(shape[:2]) % tile_shape[:2]) != 0
    ):
        pytest.xfail("HCOMPRESS requires 2D tiles.")
    random = numpy_rng.uniform(high=255, size=shape)
    # Set first value to be exactly zero as zero values require special treatment
    # for SUBTRACTIVE_DITHER_2
    random.ravel()[0] = 0.0
    original_data = random.astype(dtype)

    dtype_sanitizer = {
        ">": "big",
        "<": "little",
        "=": "native",
    }
    filename = (
        tmp_path / f"{compression_type}_{dtype[1:]}_{dtype_sanitizer[dtype[0]]}.fits"
    )

    param = fitsio_param_to_astropy_param(compression_param)
    hdu = fits.CompImageHDU(
        data=original_data,
        compression_type=compression_type,
        tile_shape=tile_shape,
        **param,
    )
    hdu.writeto(filename)

    atol = 0
    if compression_param.get("qmethod", None) is not None:
        # This is a horrific hack We are comparing quantized data to unquantized
        # data here, so there can be pretty large differences.  What this test
        # is really checking for is arrays which are *completely* different,
        # which would indicate the compression has not worked.
        atol = 17

    with fits.open(filename) as hdul:
        a = hdul[1].data
        np.testing.assert_allclose(original_data, hdul[1].data, atol=atol)


def test_plio_1_out_of_range():
    pc = PLIO1(tilesize=10)
    data = np.arange(-10, 0).astype(np.int32)

    with pytest.raises(ValueError):
        pc.encode(data)


def test_invalid_tile(tmp_path):
    # Regression test for a bug that caused a segmentation fault if the data
    # for a tile would have resulted in a warning

    m13_rice_path = Path(__file__).parent / "data" / "m13_rice.fits"

    # For this test, we will change the length of the first chunk so that it is
    # invalid. To do this, we just change the bytes directly so we first check
    # a checksum of the file to make sure it is as we expect to start with.

    with open(m13_rice_path, "rb") as f:
        content = f.read()

    assert hashlib.sha256(content).hexdigest()[:8] == "de6d2f69"

    # We change bytes 8640 to 8643 which are the length of the first tile:

    assert content[8640:8644] == b"\x00\x00\x00\x96"

    with open(tmp_path / "m13_corrupted.fits", "wb") as f:
        f.write(content[:8640])
        f.write(b"\x00\x00\x00\x95")
        f.write(content[8644:])

    with fits.open(tmp_path / "m13_corrupted.fits") as hdulist:
        # Access the data to make sure we decompress it
        with pytest.raises(
            CfitsioException,
            match="decompression error: hit end of compressed byte stream",
        ):
            hdulist[1].data.sum()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# STDLIB
import contextlib
import os
import re
from math import ceil

from astropy import online_docs_root
from astropy.io.votable import exceptions
from astropy.utils.xml.writer import XMLWriter, xml_escape

html_header = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html
        PUBLIC "-//W3C//DTD XHTML Basic 1.0//EN"
        "http://www.w3.org/TR/xhtml-basic/xhtml-basic10.dtd">
"""

default_style = """
body {
font-family: sans-serif
}
a {
text-decoration: none
}
.highlight {
color: red;
font-weight: bold;
text-decoration: underline;
}
.green { background-color: #ddffdd }
.red   { background-color: #ffdddd }
.yellow { background-color: #ffffdd }
tr:hover { background-color: #dddddd }
table {
        border-width: 1px;
        border-spacing: 0px;
        border-style: solid;
        border-color: gray;
        border-collapse: collapse;
        background-color: white;
        padding: 5px;
}
table th {
        border-width: 1px;
        padding: 5px;
        border-style: solid;
        border-color: gray;
}
table td {
        border-width: 1px;
        padding: 5px;
        border-style: solid;
        border-color: gray;
}
"""


@contextlib.contextmanager
def make_html_header(w):
    w.write(html_header)
    with w.tag("html", xmlns="http://www.w3.org/1999/xhtml", lang="en-US"):
        with w.tag("head"):
            w.element("title", "VO Validation results")
            w.element("style", default_style)

            with w.tag("body"):
                yield


def write_source_line(w, line, nchar=0):
    part1 = xml_escape(line[:nchar].decode("utf-8"))
    char = xml_escape(line[nchar : nchar + 1].decode("utf-8"))
    part2 = xml_escape(line[nchar + 1 :].decode("utf-8"))

    w.write("  ")
    w.write(part1)
    w.write(f'<span class="highlight">{char}</span>')
    w.write(part2)
    w.write("\n\n")


def write_warning(w, line, xml_lines):
    warning = exceptions.parse_vowarning(line)
    if not warning["is_something"]:
        w.data(line)
    else:
        w.write(f"Line {warning['nline']:d}: ")
        if warning["warning"]:
            w.write(
                '<a href="{}/{}">{}</a>: '.format(
                    online_docs_root, warning["doc_url"], warning["warning"]
                )
            )
        msg = warning["message"]
        if not isinstance(warning["message"], str):
            msg = msg.decode("utf-8")
        w.write(xml_escape(msg))
        w.write("\n")
        if 1 <= warning["nline"] < len(xml_lines):
            write_source_line(w, xml_lines[warning["nline"] - 1], warning["nchar"])


def write_votlint_warning(w, line, xml_lines):
    match = re.search(
        r"(WARNING|ERROR|INFO) \(l.(?P<line>[0-9]+), c.(?P<column>[0-9]+)\):"
        r" (?P<rest>.*)",
        line,
    )
    if match:
        w.write(f"Line {int(match.group('line'))}: {xml_escape(match.group('rest'))}\n")
        write_source_line(
            w, xml_lines[int(match.group("line")) - 1], int(match.group("column")) - 1
        )
    else:
        w.data(line)
        w.data("\n")


def write_result(result):
    if "network_error" in result and result["network_error"] is not None:
        return

    xml = result.get_xml_content()
    xml_lines = xml.splitlines()

    path = os.path.join(result.get_dirpath(), "index.html")

    with open(path, "w", encoding="utf-8") as fd:
        w = XMLWriter(fd)
        with make_html_header(w):
            with w.tag("p"):
                with w.tag("a", href="vo.xml"):
                    w.data(result.url.decode("ascii"))
            w.element("hr")

            with w.tag("pre"):
                w._flush()
                for line in result["warnings"]:
                    write_warning(w, line, xml_lines)

            if result["xmllint"] is False:
                w.element("hr")
                w.element("p", "xmllint results:")
                content = result["xmllint_content"]
                if not isinstance(content, str):
                    content = content.decode("ascii")
                content = content.replace(result.get_dirpath() + "/", "")
                with w.tag("pre"):
                    w.data(content)

            if "votlint" in result:
                if result["votlint"] is False:
                    w.element("hr")
                    w.element("p", "votlint results:")
                    content = result["votlint_content"]
                    if not isinstance(content, str):
                        content = content.decode("ascii")
                    with w.tag("pre"):
                        w._flush()
                        for line in content.splitlines():
                            write_votlint_warning(w, line, xml_lines)


def write_result_row(w, result):
    with w.tag("tr"):
        with w.tag("td"):
            if "network_error" in result and result["network_error"] is not None:
                w.data(result.url.decode("ascii"))
            else:
                w.element(
                    "a",
                    result.url.decode("ascii"),
                    href=f"{result.get_htmlpath()}/index.html",
                )

        if "network_error" in result and result["network_error"] is not None:
            w.element("td", str(result["network_error"]), attrib={"class": "red"})
            w.element("td", "-")
            w.element("td", "-")
            w.element("td", "-")
            w.element("td", "-")
        else:
            w.element("td", "-", attrib={"class": "green"})

            if result["nexceptions"]:
                cls = "red"
                msg = "Fatal"
            elif result["nwarnings"]:
                cls = "yellow"
                msg = str(result["nwarnings"])
            else:
                cls = "green"
                msg = "-"
            w.element("td", msg, attrib={"class": cls})

            msg = result["version"]
            if result["xmllint"] is None:
                cls = ""
            elif result["xmllint"] is False:
                cls = "red"
            else:
                cls = "green"
            w.element("td", msg, attrib={"class": cls})

            if result["expected"] == "good":
                cls = "green"
                msg = "-"
            elif result["expected"] == "broken":
                cls = "red"
                msg = "net"
            elif result["expected"] == "incorrect":
                cls = "yellow"
                msg = "invalid"
            w.element("td", msg, attrib={"class": cls})

            if "votlint" in result:
                if result["votlint"]:
                    cls = "green"
                    msg = "Passed"
                else:
                    cls = "red"
                    msg = "Failed"
            else:
                cls = ""
                msg = "?"
            w.element("td", msg, attrib={"class": cls})


def write_table(basename, name, results, root="results", chunk_size=500):
    def write_page_links(j):
        if npages <= 1:
            return
        with w.tag("center"):
            if j > 0:
                w.element("a", "<< ", href=f"{basename}_{j - 1:02d}.html")
            for i in range(npages):
                if i == j:
                    w.data(str(i + 1))
                else:
                    w.element("a", str(i + 1), href=f"{basename}_{i:02d}.html")
                w.data(" ")
            if j < npages - 1:
                w.element("a", ">>", href=f"{basename}_{j + 1:02d}.html")

    npages = int(ceil(float(len(results)) / chunk_size))

    for i, j in enumerate(range(0, max(len(results), 1), chunk_size)):
        subresults = results[j : j + chunk_size]
        path = os.path.join(root, f"{basename}_{i:02d}.html")
        with open(path, "w", encoding="utf-8") as fd:
            w = XMLWriter(fd)
            with make_html_header(w):
                write_page_links(i)

                w.element("h2", name)

                with w.tag("table"):
                    with w.tag("tr"):
                        w.element("th", "URL")
                        w.element("th", "Network")
                        w.element("th", "Warnings")
                        w.element("th", "Schema")
                        w.element("th", "Expected")
                        w.element("th", "votlint")

                    for result in subresults:
                        write_result_row(w, result)

                write_page_links(i)


def add_subset(w, basename, name, subresults, inside=["p"], total=None):
    with w.tag("tr"):
        subresults = list(subresults)
        if total is None:
            total = len(subresults)
        if total == 0:  # pragma: no cover
            percentage = 0.0
        else:
            percentage = float(len(subresults)) / total
        with w.tag("td"):
            for element in inside:
                w.start(element)
            w.element("a", name, href=f"{basename}_00.html")
            for element in reversed(inside):
                w.end(element)
        numbers = f"{len(subresults):d} ({percentage:.2%})"
        with w.tag("td"):
            w.data(numbers)


def write_index(subsets, results, root="results"):
    path = os.path.join(root, "index.html")
    with open(path, "w", encoding="utf-8") as fd:
        w = XMLWriter(fd)
        with make_html_header(w):
            w.element("h1", "VO Validation results")

            with w.tag("table"):
                for subset in subsets:
                    add_subset(w, *subset, total=len(results))


def write_index_table(
    root, basename, name, subresults, inside=None, total=None, chunk_size=500
):
    if total is None:
        total = len(subresults)
    percentage = float(len(subresults)) / total
    numbers = f"{len(subresults):d} ({percentage:.2%})"
    write_table(basename, name + " " + numbers, subresults, root, chunk_size)



#cython: language_level=3, boundscheck=False
""" Cython inverse efuncs for cosmology integrals"""

cimport cython
from libc.math cimport exp, pow

## Inverse efunc methods for various dark energy subclasses
## These take only scalar arguments since that is what the integral
## routines give them.

## Implementation notes:
##  * Using a python list for nu_y seems to be faster than a ndarray,
##     given that nu_y generally has a small number of elements,
##     even when you turn off bounds checking, etc.
##  * Using pow(x, -0.5) is slightly faster than x**(-0.5) and
##    even more so than 1.0 / sqrt(x)
##  * Hardwiring in the p, 1/p, k, prefac values in nufunc is
##       nontrivially faster than declaring them with cdef

######### LambdaCDM
# No relativistic species
def lcdm_inv_efunc_norel(double z, double Om0, double Ode0, double Ok0):
  cdef double opz = 1.0 + z
  return pow(opz**2 * (opz * Om0 + Ok0) + Ode0, -0.5)

# Massless neutrinos
def lcdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Ok0,
    double Or0):
  cdef double opz = 1.0 + z
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 + Ode0, -0.5)

# With massive neutrinos
def lcdm_inv_efunc(double z, double Om0, double Ode0, double Ok0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 + Ode0, -0.5)

######## FlatLambdaCDM
# No relativistic species
def flcdm_inv_efunc_norel(double z, double Om0, double Ode0):
  return pow((1. + z)**3 * Om0 + Ode0, -0.5)

# Massless neutrinos
def flcdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Or0):
  cdef double opz = 1.0 + z
  return pow(opz**3 * (opz * Or0 + Om0) + Ode0, -0.5)

# With massive neutrinos
def flcdm_inv_efunc(double z, double Om0, double Ode0, double Ogamma0,
    double NeffPerNu, int nmasslessnu, list nu_y):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  return pow(opz**3 * (opz * Or0 + Om0) + Ode0, -0.5)

######## wCDM
# No relativistic species
def wcdm_inv_efunc_norel(double z, double Om0, double Ode0,
    double Ok0, double w0):
  cdef double opz = 1.0 + z
  return pow(opz**2 * (opz * Om0 + Ok0) +
            Ode0 * opz**(3. * (1.0 + w0)), -0.5)

# Massless neutrinos
def wcdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Ok0,
    double Or0, double w0):
  cdef double opz = 1.0 + z
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 +
          Ode0 * opz**(3. * (1.0 + w0)), -0.5)

# With massive neutrinos
def wcdm_inv_efunc(double z, double Om0, double Ode0, double Ok0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 +
          Ode0 * opz**(3. * (1.0 + w0)), -0.5)

######## Flat wCDM
# No relativistic species
def fwcdm_inv_efunc_norel(double z, double Om0, double Ode0, double w0):
  cdef double opz = 1.0 + z
  return pow(opz**3 * Om0 + Ode0 * opz**(3. * (1.0 + w0)), -0.5)

# Massless neutrinos
def fwcdm_inv_efunc_nomnu(double z, double Om0, double Ode0,
    double Or0, double w0):
  cdef double opz = 1.0 + z
  return pow(opz**3 * (opz * Or0 + Om0) +
            Ode0 * opz**(3. * (1.0 + w0)), -0.5)

# With massive neutrinos
def fwcdm_inv_efunc(double z, double Om0, double Ode0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  return pow(opz**3 * (opz * Or0 + Om0) + Ode0 * opz**(3. * (1.0 + w0)), -0.5)

######## w0waCDM
# No relativistic species
def w0wacdm_inv_efunc_norel(double z, double Om0, double Ode0, double Ok0,
    double w0, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow(opz**2 * (opz * Om0 + Ok0) + Ode0 * Odescl, -0.5)

# Massless neutrinos
def w0wacdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Ok0,
    double Or0, double w0, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 +
          Ode0 * Odescl, -0.5)

def w0wacdm_inv_efunc(double z, double Om0, double Ode0, double Ok0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0,
    double wa):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 + Ode0 * Odescl, -0.5)

######## Flatw0waCDM
# No relativistic species
def fw0wacdm_inv_efunc_norel(double z, double Om0, double Ode0,
    double w0, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow(opz**3 * Om0 + Ode0 * Odescl, -0.5)

# Massless neutrinos
def fw0wacdm_inv_efunc_nomnu(double z, double Om0, double Ode0,
    double Or0, double w0, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

# With massive neutrinos
def fw0wacdm_inv_efunc(double z, double Om0, double Ode0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0,
    double wa):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1 + w0 + wa)) * exp(-3.0 * wa * z / opz)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

######## wpwaCDM
# No relativistic species
def wpwacdm_inv_efunc_norel(double z, double Om0, double Ode0, double Ok0,
    double wp, double apiv, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow(opz**2 * (opz * Om0 + Ok0) + Ode0 * Odescl, -0.5)

# Massless neutrinos
def wpwacdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Ok0,
    double Or0, double wp, double apiv, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 +
          Ode0 * Odescl, -0.5)

# With massive neutrinos
def wpwacdm_inv_efunc(double z, double Om0, double Ode0, double Ok0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double wp,
    double apiv, double wa):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 + Ode0 * Odescl, -0.5)

######## FlatwpwaCDM
# No relativistic species
def fwpwacdm_inv_efunc_norel(double z, double Om0, double Ode0,
    double wp, double apiv, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow(opz**3 * Om0 + Ode0 * Odescl, -0.5)

# Massless neutrinos
def fwpwacdm_inv_efunc_nomnu(double z, double Om0, double Ode0,
    double Or0, double wp, double apiv, double wa):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

# With massive neutrinos
def fwpwacdm_inv_efunc(double z, double Om0, double Ode0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double wp,
    double apiv, double wa):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1. + wp + apiv * wa)) * exp(-3. * wa * z / opz)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

######## w0wzCDM
# No relativistic species
def w0wzcdm_inv_efunc_norel(double z, double Om0, double Ode0, double Ok0,
    double w0, double wz):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow(opz**2 * (opz * Om0 + Ok0) + Ode0 * Odescl, -0.5)

# Massless neutrinos
def w0wzcdm_inv_efunc_nomnu(double z, double Om0, double Ode0, double Ok0,
    double Or0, double w0, double wz):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 +
          Ode0 * Odescl, -0.5)

# With massive neutrinos
def w0wzcdm_inv_efunc(double z, double Om0, double Ode0, double Ok0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0,
    double wz):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow((((opz * Or0 + Om0) * opz) + Ok0) * opz**2 + Ode0 * Odescl, -0.5)

######## Flatw0wzCDM
# No relativistic species
def fw0wzcdm_inv_efunc_norel(double z, double Om0, double Ode0,
    double w0, double wz):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow(opz**3 * Om0 + Ode0 * Odescl, -0.5)

# Massless neutrinos
def fw0wzcdm_inv_efunc_nomnu(double z, double Om0, double Ode0,
    double Or0, double w0, double wz):
  cdef double opz = 1.0 + z
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

# With massive neutrinos
def fw0wzcdm_inv_efunc(double z, double Om0, double Ode0,
    double Ogamma0, double NeffPerNu, int nmasslessnu, list nu_y, double w0,
    double wz):

  cdef double opz = 1.0 + z
  cdef double Or0 = Ogamma0 * (1.0 + nufunc(opz, NeffPerNu, nmasslessnu, nu_y))
  cdef double Odescl = opz**(3. * (1. + w0 - wz)) * exp(3. * wz * z)
  return pow((opz * Or0 + Om0) * opz**3 + Ode0 * Odescl, -0.5)

######## Neutrino relative density function
# Scalar equivalent to FLRW.nu_realative_density in core.py
#  Please see that for further discussion.
# This should only be called with massive neutrinos (e.g., nu_y is not empty)
# Briefly, this is just a numerical fitting function to the true relationship,
#  which is too expensive to want to evaluate directly.  The
#  constants which appear are:
#    p = 1.83  -> numerical fitting constant from Komatsu et al.
#  1/p = 0.54644... -> same constant
#    k = 0.3173 -> another fitting constant
#  7/8 (4/11)^(4/3) = 0.2271... -> fermion/boson constant for neutrino
#                                   contribution -- see any cosmology book
#  The Komatsu reference is: Komatsu et al. 2011, ApJS 192, 18
cdef nufunc(double opz, double NeffPerNu, int nmasslessnu, list nu_y):
  cdef Py_ssize_t i, N = len(nu_y)
  cdef double k = 0.3173 / opz
  cdef double rel_mass_sum = nmasslessnu
  for i in range(N):
    rel_mass_sum += pow(1.0 + (k * <double>nu_y[i])**1.83, 0.54644808743)
  return 0.22710731766 * NeffPerNu * rel_mass_sum



"""Typing module for supporting type annotations related to :mod:`~astropy.units`."""

from __future__ import annotations

__all__ = ["QuantityLike"]


from typing import TYPE_CHECKING

import numpy.typing as npt

from astropy.units import Quantity

if TYPE_CHECKING:
    from typing import TypeAlias

# Note: Quantity is technically covered by npt.ArrayLike, but we want to
# explicitly include it here so that it is clear that we are also including
# Quantity objects in the definition of QuantityLike.
QuantityLike: TypeAlias = Quantity | npt.ArrayLike
"""Type alias for a quantity-like object.

This is an object that can be converted to a :class:`~astropy.units.Quantity` object
using the :func:`~astropy.units.Quantity` constructor.

Examples
--------
We assume the following imports:

    >>> from astropy import units as u

This is a non-exhaustive list of examples of quantity-like objects:

Integers and floats:

    >>> u.Quantity(1, u.meter)
    <Quantity 1.0 m>

    >>> u.Quantity(1.0, u.meter)
    <Quantity 1.0 m>

Lists and tuples:

    >>> u.Quantity([1.0, 2.0], u.meter)
    <Quantity [1., 2.] m>

    >>> u.Quantity((1.0, 2.0), u.meter)
    <Quantity [1., 2.] m>

Numpy arrays:

    >>> u.Quantity(np.array([1.0, 2.0]), u.meter)

:class:`~astropy.units.Quantity` objects:

    >>> u.Quantity(u.Quantity(1.0, u.meter))
    <Quantity 1.0 m>

Strings:

    >>> u.Quantity('1.0 m')
    <Quantity 1.0 m>

For more examples see the :mod:`numpy.typing` definition of
:obj:`numpy.typing.ArrayLike`.
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest

from astropy.table import Column, Table
from astropy.table.table_helpers import simple_table
from astropy.utils.compat.optional_deps import HAS_BS4
from astropy.utils.data import get_pkg_data_filename

files = [
    "data/cds.dat",
    "data/ipac.dat",
    "data/daophot.dat",
    "data/latex1.tex",
    "data/simple_csv.csv",
]


if HAS_BS4:
    files.append("data/html.html")


@pytest.mark.parametrize("filename", files)
def test_read_generic(filename):
    Table.read(get_pkg_data_filename(filename), format="ascii")


def test_write_generic(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    t.write(tmp_path / "test", format="ascii")


def test_read_ipac():
    Table.read(get_pkg_data_filename("data/ipac.dat"), format="ipac")


def test_read_cds():
    Table.read(get_pkg_data_filename("data/cds.dat"), format="cds")


def test_read_dapphot():
    Table.read(get_pkg_data_filename("data/daophot.dat"), format="daophot")


def test_read_latex():
    Table.read(get_pkg_data_filename("data/latex1.tex"), format="latex")


def test_read_latex_noformat():
    Table.read(get_pkg_data_filename("data/latex1.tex"))


def test_write_latex(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.tex"
    t.write(path, format="latex")


def test_write_latex_noformat(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.tex"
    t.write(path)


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_read_html():
    Table.read(get_pkg_data_filename("data/html.html"), format="html")


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_read_html_noformat():
    Table.read(get_pkg_data_filename("data/html.html"))


def test_write_html(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.html"
    t.write(path, format="html")


def test_write_html_noformat(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.html"
    t.write(path)


def test_read_rdb():
    Table.read(get_pkg_data_filename("data/short.rdb"), format="rdb")


def test_read_rdb_noformat():
    Table.read(get_pkg_data_filename("data/short.rdb"))


def test_write_rdb(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.rdb"
    t.write(path, format="rdb")


def test_write_rdb_noformat(tmp_path):
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.rdb"
    t.write(path)


def test_read_csv():
    """If properly registered, filename should be sufficient to specify format

    #3189
    """
    Table.read(get_pkg_data_filename("data/simple_csv.csv"))


def test_write_csv(tmp_path):
    """If properly registered, filename should be sufficient to specify format

    #3189
    """
    t = Table()
    t.add_column(Column(name="a", data=[1, 2, 3]))
    t.add_column(Column(name="b", data=["a", "b", "c"]))
    path = tmp_path / "data.csv"
    t.write(path)


def test_auto_identify_ecsv(tmp_path):
    tbl = simple_table()
    tmpfile = tmp_path / "tmpFile.ecsv"
    tbl.write(tmpfile)
    tbl2 = Table.read(tmpfile)
    assert np.all(tbl == tbl2)
